#include <algorithm>
#include <src/server/QueryServer.h>
#include "QueryClient.h"
#include <netinet/tcp.h>
#include <src/Configuration.h>
#include <log/LogUtils.h>
#include <misc/memtracker.h>
#include "src/InstanceHandler.h"
#include <pipes/errors.h>
#include <misc/std_unique_ptr.h>
#include "./QueryClientConnection.h"

using namespace std;
using namespace std::chrono;
using namespace ts;
using namespace ts::server;

#if defined(TCP_CORK) && !defined(TCP_NOPUSH)
    #define TCP_NOPUSH TCP_CORK
#endif

//#define DEBUG_TRAFFIC

QueryClient::QueryClient(QueryServer* handle, int sockfd) : ConnectedClient(handle->sql, nullptr), handle(handle) {
    memtrack::allocated<QueryClient>(this);

    this->connection = new server::query::QueryClientConnection{this, sockfd};

    this->state = ClientState::INITIALIZING;
    connectedTimestamp = system_clock::now();

    this->resetEventMask();
}

bool QueryClient::initialize(std::string& error, const std::shared_ptr<ts::server::QueryClient> &cl) {
    this->_this = cl;

    this->properties()[property::CLIENT_TYPE] = ClientType::CLIENT_QUERY;
    this->properties()[property::CLIENT_TYPE_EXACT] = ClientType::CLIENT_QUERY;
    this->properties()[property::CLIENT_UNIQUE_IDENTIFIER] = "UnknownQuery";
    this->properties()[property::CLIENT_NICKNAME] = string() + "ServerQuery#" + this->getLoggingPeerIp() + "/" + to_string(this->getPeerPort());

    DatabaseHelper::assignDatabaseId(this->sql, this->getServerId(), _this.lock());

    /* may already calls handle_connection_initialized() */
    if(!this->connection->initialize(error))
        return false;
    return true;
}

QueryClient::~QueryClient() {
    memtrack::freed<QueryClient>(this);

    delete this->connection;
}

void QueryClient::handle_connection_initialized() {
    std::lock_guard lock{this->command_lock}; /* we dont want to handle anything while we're initializing */
    this->connectTimestamp = system_clock::now();
    this->properties()[property::CLIENT_LASTCONNECTED] = duration_cast<seconds>(this->connectTimestamp.time_since_epoch()).count();

    if(ts::config::query::sslMode == 1 && this->connection->connection_type() != server::query::ConnectionType::SSL_ENCRYPTED) {
        command_result error{error::failed_connection_initialisation,  "Please use a SSL encryption!"};
        this->notifyError(error);
        error.release_details();
        this->disconnect("Please us a SSL encryption for more security.\nThe server denies also all other connections!");
        return;
    }

    writeMessage(config::query::motd);
    assert(this->handle);
    if(this->handle->ip_blacklist) {
        assert(this->handle->ip_blacklist);
        if(this->handle->ip_blacklist->contains(this->remote_address)) {
            Command cmd("error");
            auto err = findError("client_login_not_permitted");
            cmd["id"] = err.errorId;
            cmd["msg"] = err.message;
            cmd["extra_msg"] = "You're not permitted to use the query interface! (Your blacklisted)";
            this->sendCommand(cmd);
            this->disconnect("blacklisted");
            return;;
        }
        if(this->handle->ip_whitelist)
            this->whitelisted = this->handle->ip_whitelist->contains(this->remote_address);
        else
            this->whitelisted = false;
        debugMessage(LOG_QUERY, "Got new query client from {}. Whitelisted: {}", this->getLoggingPeerIp(), this->whitelisted);
    }

    if(!this->whitelisted) {
        threads::MutexLock llock(this->handle->loginLock);
        if(this->handle->queryBann.count(this->getPeerIp()) > 0) {
            auto ban = this->handle->queryBann[this->getPeerIp()];
            Command cmd("error");
            auto err = findError("server_connect_banned");
            cmd["id"] = err.errorId;
            cmd["msg"] = err.message;
            cmd["extra_msg"] = "you may retry in " + to_string(duration_cast<seconds>(ban - system_clock::now()).count()) + " seconds";
            this->sendCommand(cmd);
            this->disconnect("");
        }
    }

    this->update_cached_permissions();
}

void QueryClient::handle_connection_finalized() {
    /* when this has been called there could not be any command executing! */
    //TODO: Is this statement really true?

    if(this->server) {
        {
            unique_lock channel_lock(this->server->channel_tree_lock);
            this->server->unregisterClient(_this.lock(), "disconnected", channel_lock);
        }
        this->server->groups->disableCache(this->getClientDatabaseId());
        this->server = nullptr;
    }

    if(this->handle)
        this->handle->unregisterConnection(dynamic_pointer_cast<QueryClient>(_this.lock()));
}

void QueryClient::writeMessage(const std::string& message) {
    if(this->state == ClientState::DISCONNECTED) return;
    this->connection->send_data(message);
}


bool QueryClient::disconnect(const std::string &reason) {
    if(!reason.empty()) {
        Command cmd("disconnect");
        cmd["reason"] = reason;
        this->sendCommand(cmd);
    }
    return this->close_connection(system_clock::now() + seconds(1));
}

bool QueryClient::close_connection(const std::chrono::system_clock::time_point& flushTimeout) {
    auto ownLock = dynamic_pointer_cast<QueryClient>(_this.lock());
    if(!ownLock) return false;

    this->connection->close_connection(flushTimeout);
    return true;
}

bool QueryClient::process_next_command() {
    using CommandAssembleState = server::query::CommandAssembleState;

    lock_guard clock(this->command_lock);
    if(!this->handle || this->state == ClientState::DISCONNECTED) return false;

    string command{};
    bool more_pending{false};
    switch (this->connection->next_command(command)) {
        case CommandAssembleState::MORE_COMMANDS_PENDING:
            more_pending = true;
        case CommandAssembleState::SUCCESS:
            break;
        case CommandAssembleState::NO_COMMAND_PENDING:
            return false; /* nothing to do */
    }

    if(command.empty() || command.find_first_not_of(' ') == string::npos) { //Empty command
        logTrace(LOG_QUERY, "[{}:{}] Got query idle command (Empty command or spaces)", this->getLoggingPeerIp(), this->getPeerPort());
        CMD_RESET_IDLE; //if idle time over 5 min than connection drop
        return true;
    }

    if((uint8_t) command[0] == 255) {
        string commands{};

        /* we got a telnet command here */
        while(command.size() >= 2 && (uint8_t) command[0] == 255) {
            uint8_t code = command[1];
            uint8_t option = command[2];

            if(!commands.empty())
                commands += ", ";
            commands += to_string(code) + ":" + to_string(option);
            command = command.substr(3);
        }

        logTrace(LOG_QUERY, "[{}:{}] Received telnet command(s): {}. Ignoring it.",this->getLoggingPeerIp(), this->getPeerPort(), commands);
        CMD_RESET_IDLE;
        if(command.empty())
            return true;
    }

    unique_ptr<Command> cmd;
    command_result error{};
    try {
        cmd = make_unique<Command>(Command::parse(pipes::buffer_view{(void*) command.data(), command.length()}, true, !ts::config::server::strict_ut8_mode));
    } catch(std::invalid_argument& ex) {
        logTrace(LOG_QUERY, "[{}:{}] Failed to parse command (invalid argument): {}", this->getLoggingPeerIp(), this->getPeerPort(), command);
        error = command_result{error::parameter_convert};
        goto handle_error;
    } catch(std::exception& ex) {
        logTrace(LOG_QUERY, "[{}:{}] Failed to parse command (exception: {}): {}", this->getLoggingPeerIp(), this->getPeerPort(), ex.what(), command);
        error = command_result{error::vs_critical, std::string{ex.what()}};
        goto handle_error;
    }

    try {
        this->handleCommandFull(*cmd);
    } catch(std::exception& ex) {
        error = command_result{error::vs_critical, std::string{ex.what()}};
        goto handle_error;
    }
    return more_pending;

    handle_error:
    this->notifyError(error);
    error.release_details();
    return false;
}

void QueryClient::sendCommand(const ts::Command &command, bool) {
    auto cmd = command.build();
    writeMessage(cmd + config::query::newlineCharacter);
    logTrace(LOG_QUERY, "Send command {}", cmd);
}

void QueryClient::sendCommand(const ts::command_builder &command, bool) {
    writeMessage(command.build() + config::query::newlineCharacter);
    logTrace(LOG_QUERY, "Send command {}", command.build());
}

void QueryClient::tick(const std::chrono::system_clock::time_point &time) {
    ConnectedClient::tick(time);
}

void QueryClient::queryTick() {
    if(this->idleTimestamp.time_since_epoch().count() > 0 && system_clock::now() - this->idleTimestamp > minutes(5)){
        debugMessage(LOG_QUERY, "Dropping client " + this->getLoggingPeerIp() + "|" + this->getDisplayName() + ". (Timeout)");
        this->close_connection(system_clock::now() + seconds(1));
    }

    if(this->connection->connection_type() == server::query::ConnectionType::UNKNOWN && system_clock::now() - milliseconds{500} > connectedTimestamp) {
        this->connection->enforce_text_connection();
    }
}

bool QueryClient::notifyChannelSubscribed(const deque<shared_ptr<BasicChannel>> &) {
    return false;
}

bool QueryClient::notifyChannelUnsubscribed(const deque<shared_ptr<BasicChannel>> &){
    return false;
}

bool QueryClient::ignoresFlood() {
    return this->whitelisted || ConnectedClient::ignoresFlood();
}

void QueryClient::disconnect_from_virtual_server() {
    threads::MutexLock lock(this->command_lock);

    auto server_locked = this->server;
    if(server_locked) {
        //unregister manager from old server
        {
            unique_lock tree_lock(this->server->channel_tree_lock);
            if(this->currentChannel)
                this->server->client_move(this->ref(), nullptr, nullptr, "", ViewReasonId::VREASON_USER_ACTION, false, tree_lock);
            this->server->unregisterClient(_this.lock(), "server switch", tree_lock);
        }
        server_locked->groups->disableCache(this->getClientDatabaseId());
        this->channels->reset();
        this->currentChannel = nullptr;
        this->server = nullptr;
        this->loadDataForCurrentServer();
    }
    serverInstance->getGroupManager()->enableCache(this->getClientDatabaseId());
}