#pragma once

#include <misc/spin_lock.h>
#include <mutex>
#include <deque>
#include <protocol/Packet.h>
#include <protocol/generation.h>
#include <protocol/ringbuffer.h>

namespace ts::connection {
    class CryptHandler;
    class CompressionHandler;
    class AcknowledgeManager;
}

namespace ts::stats {
    class ConnectionStatistics;
}

namespace ts::server::server::udp {
    struct CommandFragment {
        uint16_t packet_id{0};
        uint16_t packet_generation{0};

        uint8_t packet_flags{0};
        uint32_t payload_length : 24;
        pipes::buffer payload{};

        CommandFragment() { this->payload_length = 0; }
        CommandFragment(uint16_t packetId, uint16_t packetGeneration, uint8_t packetFlags, uint32_t payloadLength, pipes::buffer payload)
                : packet_id{packetId}, packet_generation{packetGeneration}, packet_flags{packetFlags}, payload_length{payloadLength}, payload{std::move(payload)} {}

        CommandFragment& operator=(const CommandFragment&) = default;
        CommandFragment(const CommandFragment& other) = default;
        CommandFragment(CommandFragment&&) = default;
    };
    static_assert(sizeof(CommandFragment) == 8 + sizeof(pipes::buffer));

    enum struct PacketDecodeResult {
        SUCCESS,

        INVALID_PACKET,
        DUPLICATED_PACKET, /* error message contains debug properties */

        BUFFER_OVERFLOW, /* error message contains debug properties */

        DECRYPT_KEY_GEN_FAILED,
        DECRYPT_FAILED, /* has custom message */

        COMMAND_INSTERT_FAILED
    };

    enum struct CommandReassembleResult {
        SUCCESS,

        MORE_COMMANDS_PENDING, /* equal with success */
        NO_COMMANDS_PENDING,

        COMMAND_TOO_LARGE, /* this is a fatal error to the connection */
        COMMAND_DECOMPRESS_FAILED,

        SEQUENCE_LENGTH_TOO_LONG /* unrecoverable error */
    };

    class PacketDecoder {
            typedef protocol::PacketRingBuffer<CommandFragment, 32, CommandFragment> command_fragment_buffer_t;
            typedef std::array<command_fragment_buffer_t, 2> command_packet_reassembler;
        public:
            //typedef std::function<void(const protocol::ClientPacketParser&)> callback_decoded_packet_t;
            //typedef std::function<void(uint16_t /* packet id */, bool /* is command low */)> callback_send_acknowledge_t;
            //typedef std::function<void(PacketDecodeResult /* error */, const std::string& /* custom message */)> callback_decode_failed_t;
            /* gets better optimized out */
            typedef void(*callback_decoded_packet_t)(void* /* cb argument */, const protocol::ClientPacketParser&);
            typedef void(*callback_send_acknowledge_t)(void* /* cb argument */, uint16_t /* packet id */, bool /* is command low */);
            typedef void(*callback_decode_failed_t)(void* /* cb argument */, PacketDecodeResult /* error */, const std::string& /* custom message */);

            explicit PacketDecoder(connection::CryptHandler* /* crypt handler */);
            ~PacketDecoder();

            void reset();

            bool verify_encryption(const pipes::buffer_view& /* full packet */);

            /* true if commands might be pending */
            bool decode_incoming_data(const pipes::buffer_view &/* buffer */);

            /* This method is not thread save! Only one concurrent call supported */
            CommandReassembleResult reassemble_command(pipes::buffer& /* result */, bool& /* is command low */);
            void force_insert_command(const pipes::buffer_view& /* payload */);

            void register_initiv_packet();

            [[nodiscard]] inline std::shared_ptr<stats::ConnectionStatistics> get_statistics() { return this->statistics_; }
            inline void set_statistics(const std::shared_ptr<stats::ConnectionStatistics>& stats) { this->statistics_ = stats; }

            [[nodiscard]] inline bool is_protocol_encrypted() const { return this->protocol_encrypted; }
            void set_protocol_encrypted(bool flag) { this->protocol_encrypted = flag; }

            void* callback_argument{nullptr};
            callback_decoded_packet_t callback_decoded_packet{[](auto, auto&){}}; /* needs to be valid all the time! */
            callback_send_acknowledge_t callback_send_acknowledge{[](auto, auto, auto){}}; /* needs to be valid all the time! */
            callback_decode_failed_t callback_decode_failed{nullptr};
        private:
            bool protocol_encrypted{false};
            std::shared_ptr<stats::ConnectionStatistics> statistics_{nullptr};

            connection::CryptHandler* crypt_handler_{nullptr};

            std::array<protocol::generation_estimator, 9> incoming_generation_estimators{}; /* implementation is thread save */

            std::recursive_mutex packet_buffer_lock;
            command_packet_reassembler _command_fragment_buffers;

            static inline uint8_t command_fragment_buffer_index(uint8_t packet_index) {
                return packet_index & 0x1U; /* use 0 for command and 1 for command low */
            }

            PacketDecodeResult decode_incoming_data_(std::string& /* error */, bool& /* needs command reassemble */, const pipes::buffer_view &/* buffer */);
    };
}