//
// Created by WolverinDEV on 09/03/2020.
//
#include "PacketEncoder.h"

#include <protocol/buffers.h>
#include <protocol/AcknowledgeManager.h>
#include <protocol/CompressionHandler.h>
#include <protocol/CryptHandler.cpp>

#include "../../ConnectionStatistics.h"

using namespace ts;
using namespace ts::server::server::udp;

PacketEncoder::PacketEncoder(ts::connection::CryptHandler *crypt_handler, ts::connection::CompressionHandler *compress_handler,
                             ts::connection::AcknowledgeManager *ack_handler) : crypt_handler_{crypt_handler}, compress_handler_{compress_handler}, acknowledge_handler_{ack_handler} {
    memtrack::allocated<PacketEncoder>(this);
}

PacketEncoder::~PacketEncoder() {
    memtrack::freed<PacketEncoder>(this);
    this->reset();
}

void PacketEncoder::reset() {
    for(auto& category : this->write_preprocess_queues) {
        std::lock_guard work_lock{category.work_lock};
        std::lock_guard queue_lock{category.queue_lock};

        category.queue.clear();
    }
    this->id_generator_.reset();
}

bool PacketEncoder::encode_packet(const std::shared_ptr<protocol::ServerPacket> &original_packet, EncodeFlags flags) {
    std::shared_ptr<protocol::ServerPacket> packet;
    if(flags & EncodeFlags::no_copy) {
        packet = original_packet;
    } else {
        packet = protocol::ServerPacket::from_buffer(original_packet->buffer().dup(buffer::allocate_buffer(original_packet->buffer().length())));
        if(original_packet->getListener())
            packet->setListener(std::move(original_packet->getListener()));
        packet->memory_state.flags = original_packet->memory_state.flags;
    }

    auto type = EncodeProcessCategory::from_type(packet->type().type());
    auto& queue = this->write_preprocess_queues[type];

    if(flags & EncodeFlags::sync) {
        std::string error{};
        std::vector<pipes::buffer> buffers{};
        this->process_count++;

        {
            std::unique_lock work_lock{queue.work_lock};
            auto encode_result = this->encode_packet_(error, buffers, packet, work_lock);
            if(encode_result != PacketEncodeResult::SUCCESS) {
                if(auto callback{this->callback_encode_failed}; callback)
                    callback(this->callback_argument, original_packet, encode_result, error);
                goto sync_cleanup_exit;
            }
        }

        if(auto callback{this->callback_encoded_buffers}; callback)
            callback(this->callback_argument, buffers);

        sync_cleanup_exit:
        this->process_count--; /* we're now done preparing */
        return false;
    } else {
        std::lock_guard queue_lock{queue.queue_lock};
        queue.queue.push_back(packet);
        queue.has_packets = true;
    }

    return true;
}

bool PacketEncoder::do_encode() {
    std::string error{};
    std::vector<pipes::buffer> buffers{};
    std::shared_ptr<protocol::ServerPacket> packet{nullptr};
    bool flag_more{false};

    this->process_count++; /* we're preparing a packet */
    for(auto& category : this->write_preprocess_queues) {
        if(!category.has_packets) {
            continue;
        } else if(packet) {
            flag_more = true;
            break;
        }

        std::unique_lock work_lock{category.work_lock, std::try_to_lock};
        if(!work_lock) continue; /* This particular category will already be processed */

        {
            std::lock_guard buffer_lock{category.queue_lock};
            if(category.queue.empty()) {
                category.has_packets = false;
                continue;
            }

            packet = std::move(category.queue.front());
            category.queue.pop_front();
            category.has_packets = !category.queue.empty();
            flag_more = category.has_packets;
        }

        if(auto errc = this->encode_packet_(error, buffers, packet, work_lock); errc != PacketEncodeResult::SUCCESS) {
            if(auto callback{this->callback_encode_failed}; callback)
                callback(this->callback_argument, packet, errc, error);
            if(flag_more)
                break;
            else
                continue; /* find out if we have more */
        }

        if(flag_more)
            break;
        else
            continue; /* find out if we have more */
    }

    /* enqueue buffers for write */
    if(!buffers.empty()) {
        if(auto callback{this->callback_encoded_buffers}; callback)
            callback(this->callback_argument, buffers);
    }
    this->process_count--; /* we're now done preparing */

    return flag_more;
}

PacketEncodeResult PacketEncoder::encode_packet_(std::string& error,
                                                 std::vector<pipes::buffer> &result,
                                                 const std::shared_ptr<protocol::ServerPacket> &packet,
                                                 std::unique_lock<std::mutex> &work_lock) {
    assert(work_lock.owns_lock());

    if(packet->type().compressable() && !packet->memory_state.fragment_entry) {
        packet->enable_flag(PacketFlag::Compressed);
        if(!this->compress_handler_->progressPacketOut(&*packet, error))
            return PacketEncodeResult::COMPRESS_FAILED;
    }

    std::vector<shared_ptr<ServerPacket>> fragments;
    fragments.reserve((size_t) (packet->data().length() / packet->type().max_length()) + 1);

    if(packet->data().length() > packet->type().max_length()) {
        if(!packet->type().fragmentable())
            return PacketEncodeResult::PACKET_TOO_LARGE;

        { //Split packets
            auto buffer = packet->data();

            const auto max_length = packet->type().max_length();
            while(buffer.length() > max_length * 2) {
                fragments.push_back(make_shared<ServerPacket>(packet->type(), buffer.view(0, max_length).dup(buffer::allocate_buffer(max_length))));
                buffer = buffer.range((size_t) max_length);
            }

            if(buffer.length() > max_length) { //Divide rest by 2
                fragments.push_back(make_shared<ServerPacket>(packet->type(), buffer.view(0, buffer.length() / 2).dup(buffer::allocate_buffer(buffer.length() / 2))));
                buffer = buffer.range(buffer.length() / 2);
            }
            fragments.push_back(make_shared<ServerPacket>(packet->type(), buffer));

            for(const auto& frag : fragments) {
                frag->setFragmentedEntry(true);
                frag->enable_flag(PacketFlag::NewProtocol);
            }
        }

        assert(fragments.size() >= 2);
        fragments.front()->enable_flag(PacketFlag::Fragmented);
        if(packet->has_flag(PacketFlag::Compressed))
            fragments.front()->enable_flag(PacketFlag::Compressed);

        fragments.back()->enable_flag(PacketFlag::Fragmented);

        if(packet->getListener())
            fragments.back()->setListener(std::move(packet->getListener())); //Move the listener to the last :)
    } else {
        fragments.push_back(packet);
    }

    result.reserve(fragments.size());

    /* apply packet ids */
    for(const auto& fragment : fragments) {
        if(!fragment->memory_state.id_branded)
            fragment->applyPacketId(this->id_generator_);
    }
    work_lock.unlock(); /* the rest could be unordered */


    CryptHandler::key_t crypt_key{};
    CryptHandler::nonce_t crypt_nonce{};

    auto statistics = this->statistics_;
    for(const auto& fragment : fragments) {
        if(fragment->has_flag(PacketFlag::Unencrypted)) {
            this->crypt_handler_->write_default_mac(fragment->mac().data_ptr());
        } else {
            if(this->protocol_encrypted) {
                if(!this->crypt_handler_->generate_key_nonce(false, fragment->type().type(), fragment->packetId(), fragment->generationId(), crypt_key, crypt_nonce))
                    return PacketEncodeResult::ENCRYPT_KEY_GEN_FAILED;
            } else {
                crypt_key = CryptHandler::default_key;
                crypt_nonce = CryptHandler::default_nonce;
            }

            auto crypt_result = this->crypt_handler_->encrypt(fragment->header().data_ptr(), fragment->header().length(),
                                                            fragment->data().data_ptr(), fragment->data().length(),
                                                            fragment->mac().data_ptr(),
                                                            crypt_key, crypt_nonce, error);
            if(!crypt_result)
                return PacketEncodeResult::ENCRYPT_FAILED;
        }

        if(statistics)
            statistics->logOutgoingPacket(*fragment);
        this->acknowledge_handler_->process_packet(*fragment);
        result.push_back(fragment->buffer());
    }

    return PacketEncodeResult::SUCCESS;
}