#pragma once

#include <misc/spin_lock.h>
#include <mutex>
#include <deque>
#include <protocol/Packet.h>

namespace ts::connection {
    class CryptHandler;
    class CompressionHandler;
    class AcknowledgeManager;
}

namespace ts::stats {
    class ConnectionStatistics;
}

namespace ts::server::server::udp {
    struct EncodeProcessCategory {
        enum value {
            PING_PONG = 0, //Ping/Pongs
            ACK = 2,
            VOICE_WHISPER = 1, //Voice/Whisper
            COMMAND = 3,
            INIT = 4,

            MAX = INIT
        };

        inline static value from_type(protocol::PacketType type) {
            switch(type) {
                case protocol::PING:
                case protocol::PONG:
                    return value::PING_PONG;

                case protocol::VOICE:
                case protocol::VOICE_WHISPER:
                    return value::VOICE_WHISPER;

                case protocol::ACK:
                case protocol::ACK_LOW:
                    return value::ACK;

                case protocol::COMMAND:
                case protocol::COMMAND_LOW:
                    return value::COMMAND;

                default:
                    return value::INIT;
            }
        }
    };

    enum struct PacketEncodeResult {
        SUCCESS,

        COMPRESS_FAILED, /* has custom message */
        PACKET_TOO_LARGE,
        ENCRYPT_KEY_GEN_FAILED,
        ENCRYPT_FAILED, /* has custom message */
    };

    class PacketEncoder {
        public:
            enum EncodeFlags {
                none = 0x0,
                no_copy = 0x1, /* do not copy the packet */
                sync = 0x02 /* directly process the packet */
            };

            //typedef std::function<void(const std::vector<pipes::buffer>& /* buffers */)> callback_encoded_buffers_t;
            //typedef std::function<void(const std::shared_ptr<protocol::ServerPacket> &/* the packet */, PacketEncodeResult /* error */, std::string& /* custom message */)> callback_encode_failed_t;
            /* gets better optimized out */
            typedef void(*callback_encoded_buffers_t)(void* /* cb argument */, const std::vector<pipes::buffer>& /* buffers */);
            typedef void(*callback_encode_failed_t)(void* /* cb argument */, const std::shared_ptr<protocol::ServerPacket> &/* the packet */, PacketEncodeResult /* error */, const std::string& /* custom message */);

            PacketEncoder(connection::CryptHandler* /* crypt handler */, connection::CompressionHandler* /* compress handler */, connection::AcknowledgeManager* /* acknowledge handler */);
            ~PacketEncoder();

            void reset();

            /* returns true if the encoder has something to encode */
            bool encode_packet(const std::shared_ptr<protocol::ServerPacket> &/* the packet */, EncodeFlags /* flags */);
            bool do_encode();

            [[nodiscard]] inline protocol::PacketIdManager& id_generator() { return this->id_generator_; }

            [[nodiscard]] inline std::shared_ptr<stats::ConnectionStatistics> get_statistics() { return this->statistics_; }
            inline void set_statistics(const std::shared_ptr<stats::ConnectionStatistics>& stats) { this->statistics_ = stats; }

            [[nodiscard]] inline bool is_protocol_encrypted() const { return this->protocol_encrypted; }
            void set_protocol_encrypted(bool flag) { this->protocol_encrypted = flag; }

            void* callback_argument{nullptr};
            callback_encoded_buffers_t callback_encoded_buffers{};
            callback_encode_failed_t callback_encode_failed{};
        private:
            bool protocol_encrypted{false};
            std::shared_ptr<stats::ConnectionStatistics> statistics_{nullptr};

            connection::CryptHandler* crypt_handler_{nullptr};
            connection::CompressionHandler* compress_handler_{nullptr};
            connection::AcknowledgeManager* acknowledge_handler_{nullptr};

            struct PacketEncodeQueue {
                bool has_packets{false};
                std::mutex work_lock{};

                spin_lock queue_lock{};
                std::deque<std::shared_ptr<protocol::ServerPacket>> queue{};
            };
            std::array<PacketEncodeQueue, EncodeProcessCategory::MAX> write_preprocess_queues{};

            /* ---------- Processing ---------- */
            /* automatically locked because packets of the same kind should be lock their "work_lock" from their WritePreprocessQueue object */
            protocol::PacketIdManager id_generator_{};

            std::atomic<size_t> process_count{0};

            PacketEncodeResult encode_packet_(std::string& /* error */, std::vector<pipes::buffer> &result/* buffers which need to be transferred */, const std::shared_ptr<protocol::ServerPacket> &packet/* the packet */, std::unique_lock<std::mutex> &work_lock /* work lock */);
    };
}