#include <memory>
#include <tomcrypt.h>
#include <arpa/inet.h>
#include <ThreadPool/Timer.h>
#include <misc/memtracker.h>
#include <log/LogUtils.h>

#include "VoiceClient.h"
#include "src/VirtualServer.h"

using namespace std;
using namespace std::chrono;
using namespace ts::server;
using namespace ts::protocol;

VoiceClient::VoiceClient(const std::shared_ptr<VirtualServer>& server, const sockaddr_storage* address) : SpeakingClient(server->sql, server) {
    assert(address);
    memtrack::allocated<VoiceClient>(this);
    memcpy(&this->remote_address, address, sizeof(sockaddr_storage));

    debugMessage(this->server->getServerId(), " Creating VoiceClient instance at {}", (void*) this);
}

VoiceClient::~VoiceClient() {
    debugMessage(this->getServerId(), " Deleting VoiceClient instance at {}", (void*) this);

    this->state = ClientState::DISCONNECTED;
    memtrack::freed<VoiceClient>(this);
}

void VoiceClient::initialize(const std::shared_ptr<connection::VoiceClientConnection> &connection) {
    assert(connection);
    this->connection_ = connection;

    this->properties()[property::CLIENT_TYPE] = ClientType::CLIENT_TEAMSPEAK;
    this->properties()[property::CLIENT_TYPE_EXACT] = ClientType::CLIENT_TEAMSPEAK;

    this->state = ClientState::INITIALIZING;
}

void VoiceClient::sendCommand0(const std::string_view& cmd, bool low, bool direct, std::unique_ptr<threads::Future<bool>> listener) {
    if(cmd.empty()) {
        logCritical(this->getServerId(), "{} Attempted to send an empty command!", CLIENT_STR_LOG_PREFIX);
        return;
    }

    auto packet = make_shared<protocol::ServerPacket>(
            low ? protocol::PacketTypeInfo::CommandLow : protocol::PacketTypeInfo::Command,
            pipes::buffer_view{(void*) cmd.data(), cmd.length()}
    );
    if(low) {
        packet->enable_flag(protocol::PacketFlag::NewProtocol);
    }
    packet->setListener(std::move(listener));

    if(auto connection{this->connection_}; connection)
        connection->send_packet(packet, false, direct);

#ifdef PKT_LOG_CMD
    logTrace(this->getServerId(), "{}[Command][Server -> Client] Sending command {}. Command low: {}. Full command: {}", CLIENT_STR_LOG_PREFIX, cmd.substr(0, cmd.find(' ')), low, cmd);
#endif
}

void VoiceClient::tick(const std::chrono::system_clock::time_point &time) {
    SpeakingClient::tick(time);

    {
        ALARM_TIMER(A1, "VoiceClient::tick", milliseconds(3));
        //TODO!
        if(this->state == ClientState::INITIALIZING) {
            if(this->last_packet_handshake.time_since_epoch().count() != 0) {
                if(time - this->last_packet_handshake > seconds(5)) {
                    debugMessage(this->getServerId(), "{} Got initialize timeout. {}. Time: {}", CLIENT_STR_LOG_PREFIX,
                                 this->getLoggingPeerIp() + ":" + to_string(this->getPeerPort()),
                                 duration_cast<seconds>(time - this->last_packet_handshake).count()
                    );
                    this->close_connection(system_clock::now() + seconds(1));
                }
            }
        }
    }
}

bool VoiceClient::disconnect(const std::string &reason) {
    return this->disconnect(VREASON_SERVER_KICK, reason, this->server->serverRoot, true);
}

bool VoiceClient::disconnect(ts::ViewReasonId reason_id, const std::string &reason, const std::shared_ptr<ts::server::ConnectedClient>& invoker, bool notify_viewer) {
    /*
     * We don't have to lock the disconnect lock here, because we're not really unregistering the client.
     * Its only for the clients own flavour and everything which the client receives after will be ignored :)
     */

    ClientState old_state{};
    {
        std::lock_guard state_lock{this->state_lock};
        if(this->state == ClientState::DISCONNECTED)
            return false; //Already disconnecting/disconnected

        old_state = this->state;
        this->state = ClientState::DISCONNECTED;
    }

    if(old_state == ClientState::CONNECTED) {
        /* Client has been successfully initialized; Send normal disconnect. */

        Command cmd("notifyclientleftview");
        cmd["reasonmsg"] = reason;
        cmd["reasonid"] = reason_id;
        cmd["clid"] = this->getClientId();
        cmd["cfid"] = this->currentChannel ? this->currentChannel->channelId() : 0; //Failed when cid = 0????
        cmd["ctid"] = 0;

        if (invoker) {
            cmd["invokerid"] = invoker->getClientId();
            cmd["invokername"] = invoker->getDisplayName();
            cmd["invokeruid"] = invoker->getUid();
        }

        if(notify_viewer && this->server) {
            unique_lock channel_lock(this->server->channel_tree_lock);
            this->server->client_move(this->ref(), nullptr, invoker, reason, reason_id, false, channel_lock);
        } else {
            threads::MutexLock lock(this->command_lock);
            auto server_channel = dynamic_pointer_cast<ServerChannel>(this->currentChannel);
            if(server_channel)
                server_channel->unregister_client(_this.lock());
            this->currentChannel = nullptr;
        }

        auto listener = make_unique<threads::Future<bool>>();
        auto weak_self = this->_this;
        listener->waitAndGetLater([weak_self](bool* success) {
            if(weak_self.expired()) return;
            auto self = weak_self.lock();
            if(!self) return;

            if(!success || !*success) {
                debugMessage(self->getServerId(), "{} Failed to receive disconnect acknowledge!", CLIENT_STR_LOG_PREFIX_(self));
            } else
                debugMessage(self->getServerId(), "{} Received disconnect acknowledge!", CLIENT_STR_LOG_PREFIX_(self));

            self->close_connection(chrono::system_clock::time_point{}); /* we received the ack, we do not need to flush anything */
        }, system_clock::now() + seconds(5));
        this->sendCommand0(cmd.build(), false, false, std::move(listener));
    } else {
        //TODO: Extra case for INIT_HIGH?
        this->close_connection(chrono::system_clock::now() + chrono::seconds{5});
    }
    return true;
}

bool VoiceClient::close_connection(const system_clock::time_point &timeout) {
    {
        std::lock_guard state_lock{this->state_lock};
        this->state = ClientState::DISCONNECTED;
    }

    auto connection = std::exchange(this->connection_, nullptr);
    if(connection) connection->close_connection(timeout);
    return true;
}

void VoiceClient::finalize() {
    /* there could not happen any IO while we're doing finalize()! */
    auto ownLock = dynamic_pointer_cast<VoiceClient>(_this.lock());
    assert(ownLock);

#if 0
    if(this->final_disconnected) {
        logError(this->getServerId(), "Tried to final disconnect {}/{} twice", this->getLoggingPeerIp() + ":" + to_string(this->getPeerPort()), this->getDisplayName());
        return;
    }
#endif
    this->state = ClientState::DISCONNECTED;

    threads::MutexLock command_lock(this->command_lock); //We should not progress any commands while disconnecting
    //Unload manager cache
    this->processLeave();
}

void VoiceClient::send_voice_packet(const pipes::buffer_view &voice_buffer, const SpeakingClient::VoicePacketFlags &flags) {
    auto packet = make_shared<ServerPacket>(PacketTypeInfo::Voice, voice_buffer.length());
    {
        PacketFlag::PacketFlags packet_flags = PacketFlag::None;
        packet_flags |= flags.encrypted ? 0U : PacketFlag::Unencrypted;
        packet_flags |= flags.head ? PacketFlag::Compressed : 0U;
        packet_flags |= flags.fragmented ? PacketFlag::Fragmented : 0U;
        packet_flags |= flags.new_protocol ? PacketFlag::NewProtocol : 0U;
        packet->set_flags(packet_flags);
    }

    memcpy(packet->data().data_ptr<void>(), voice_buffer.data_ptr<void>(), voice_buffer.length());
    if(auto connection{this->connection_}; connection)
        connection->send_packet(packet, false, false);
}

void VoiceClient::send_voice_whisper_packet(const pipes::buffer_view &voice_buffer, const SpeakingClient::VoicePacketFlags &flags) {
    auto packet = make_shared<ServerPacket>(PacketTypeInfo::VoiceWhisper, voice_buffer.length());
    {
        PacketFlag::PacketFlags packet_flags = PacketFlag::None;
        packet_flags |= flags.encrypted ? 0U : PacketFlag::Unencrypted;
        packet_flags |= flags.head ? PacketFlag::Compressed : 0U;
        packet_flags |= flags.fragmented ? PacketFlag::Fragmented : 0U;
        packet_flags |= flags.new_protocol ? PacketFlag::NewProtocol : 0U;
        packet->set_flags(packet_flags);
    }

    memcpy(packet->data().data_ptr<void>(), voice_buffer.data_ptr<void>(), voice_buffer.length());
    if(auto connection{this->connection_}; connection)
        connection->send_packet(packet, false, false);
}

std::chrono::milliseconds VoiceClient::calculatePing() {
    auto connection = this->connection_;
    return connection ? connection->ping_handler_.current_ping() : std::chrono::milliseconds{0};
}