//
// Created by WolverinDEV on 03/03/2020.
//
#include <log/LogUtils.h>
#include "./GroupAssignmentManager.h"
#include "./GroupManager.h"
#include "../vserver/VirtualServer.h"
#include "../client/ConnectedClient.h"

using namespace ts::server::groups;

GroupAssignmentManager::GroupAssignmentManager(GroupManager* handle) : manager_{handle} { }
GroupAssignmentManager::~GroupAssignmentManager() = default;

bool GroupAssignmentManager::initialize(std::string &error) {
    return true;
}

sql::SqlManager* GroupAssignmentManager::sql_manager() {
    return this->manager_->virtual_server()->sql_manager();
}

ts::ServerId GroupAssignmentManager::server_id() {
    return this->manager_->virtual_server()->server_id();
}

bool GroupAssignmentManager::load_data(std::string &error) {
    if constexpr(kCacheAllClients) {
        std::lock_guard cache_lock{this->client_cache_lock};
        std::unique_ptr<ClientCache> current_entry{nullptr};

        auto res = sql::command(this->sql_manager(), "SELECT `groupId`, `cldbid`, `channelId`, `until` FROM `assignedGroups` WHERE `serverId` = :sid ORDER BY `cldbid`", variable{":sid", this->server_id()})
                .query([&](int length, std::string* value, std::string* column) {
                    ChannelId channel_id{0};
                    GroupId group_id{0};
                    ClientDbId client_dbid{0};

                    for(int index = 0; index < length; index++){
                        try {
                            if(column[index] == "groupId"){
                                group_id = stoll(value[index]);
                            } else if(column[index] == "until"){
                            } else if(column[index] == "channelId"){
                                channel_id = stoll(value[index]);
                            } else if(column[index] == "cldbid"){
                                client_dbid = stoll(value[index]);
                            }
                        } catch(std::exception& ex) {
                            logError(this->server_id(), "Failed to load group assignment from database. Column {} contains an invalid value: {}", column[index], value[index]);
                            return 0;
                        }
                    }

                    if(current_entry)
                        if(current_entry->client_database_id != client_dbid)
                            this->client_cache.push_back(std::move(current_entry));

                    if(!current_entry) {
                        current_entry = std::make_unique<ClientCache>();
                        current_entry->client_database_id = client_dbid;
                    }

                    if(channel_id)
                        current_entry->channel_group_assignments.emplace_back(channel_id, group_id, false);
                    else
                        current_entry->server_group_assignments.emplace_back(group_id);
                    return 0;
                });
        if(!res) {
            error = "failed to query database (" + res.fmtStr() + ")";
            return false;
        }
        if(current_entry)
            this->client_cache.push_back(std::move(current_entry));
    }
    return true;
}

void GroupAssignmentManager::unload_data() {
    std::lock_guard cache_lock{this->client_cache_lock};
    this->client_cache.clear();
}

void GroupAssignmentManager::enable_cache_for_client(GroupAssignmentCalculateMode mode, ClientDbId cldbid) {
    if constexpr(!kCacheAllClients) {
        bool cache_exists{false};
        {
            std::lock_guard cache_lock{this->client_cache_lock};
            for(auto& client : this->client_cache)
                if(client->client_database_id == cldbid) {
                    client->use_count++;
                    cache_exists = true;
                    break;
                }
        }

        if(!cache_exists) {
            auto cache = std::make_unique<ClientCache>();
            cache->client_database_id = cldbid;
            cache->use_count++;

            auto res = sql::command(this->sql_manager(), "SELECT `groupId`, `channelId`, `until` FROM `assignedGroups` WHERE `serverId` = :sid AND `cldbid` = :cldbid", variable{":sid", this->server_id()}, variable{":cldbid", cldbid})
                    .query([&](int length, std::string* value, std::string* column) {
                        ChannelId channel_id{0};
                        GroupId group_id{0};

                        for(int index = 0; index < length; index++){
                            try {
                                if(column[index] == "groupId"){
                                    group_id = stoll(value[index]);
                                } else if(column[index] == "until"){
                                } else if(column[index] == "channelId"){
                                    channel_id = stoll(value[index]);
                                }
                            } catch(std::exception& ex) {
                                logError(this->server_id(), "Failed to load group assignment from database for client {}. Column {} contains an invalid value: {}", cldbid, column[index], value[index]);
                                return 0;
                            }
                        }
                        if(!group_id)
                            return 0;

                        if(channel_id)
                            cache->channel_group_assignments.emplace_back(channel_id, group_id, false);
                        else
                            cache->server_group_assignments.emplace_back(group_id);
                        return 0;
                    });

            std::lock_guard cache_lock{this->client_cache_lock};
#if 0 /* lets have some performance over double entries :D */
            for(auto& client : this->client_cache)
                if(client->client_database_id == cldbid) {
                    /* somebody already inserted that client while we've loaded him */
                    cache_exists = true;
                    break;
                }
            if(!cache_exists)
#endif
            this->client_cache.push_back(std::move(cache));
        }
    }

    if(mode == GroupAssignmentCalculateMode::GLOBAL)
        if(auto parent = this->manager_->parent_manager(); parent)
            parent->assignments().enable_cache_for_client(mode, cldbid);
}

void GroupAssignmentManager::disable_cache_for_client(GroupAssignmentCalculateMode mode, ClientDbId cldbid) {
    if constexpr(!kCacheAllClients) {
        std::lock_guard cache_lock{this->client_cache_lock};
        this->client_cache.erase(std::remove_if(this->client_cache.begin(), this->client_cache.end(), [cldbid](const std::unique_ptr<ClientCache>& client) {
            return client->client_database_id == cldbid;
        }), this->client_cache.end());
    }

    if(mode == GroupAssignmentCalculateMode::GLOBAL)
        if(auto parent = this->manager_->parent_manager(); parent)
            parent->assignments().disable_cache_for_client(mode, cldbid);
}


std::vector<ts::GroupId> GroupAssignmentManager::server_groups_of_client(ts::server::groups::GroupAssignmentCalculateMode mode,
                                                                    ts::ClientDbId cldbid) {
    std::vector<ts::GroupId> result{};
    bool cache_found{false};
    {
        std::lock_guard cache_lock{this->client_cache_lock};
        for(auto& entry : this->client_cache) {
            if(entry->client_database_id != cldbid) continue;

            result.reserve(entry->server_group_assignments.size());
            for(auto& assignment : entry->server_group_assignments)
                result.push_back(assignment.group_id);

            cache_found = true;
            break;
        }
    }

    if(!cache_found && !kCacheAllClients) {
        debugMessage(this->server_id(), "Query client groups for client {} on server {}.", cldbid, this->server_id());

        result.reserve(64);
        auto res = sql::command(this->sql_manager(), "SELECT `groupId`, `until` FROM `assignedGroups` WHERE `serverId` = :sid AND `cldbid` = :cldbid AND `channelId` = 0", variable{":sid", this->server_id()}, variable{":cldbid", cldbid})
                .query([&](int length, std::string* value, std::string* column) {
            GroupId group_id{0};
            try {
                for(int index = 0; index < length; index++) {
                    if(column[index] == "groupId") {
                        group_id = std::stoull(value[index]);
                        break;
                    }
                }
            } catch(std::exception& ex) {
                logWarning(this->server_id(), "Invalid data found in group assignment table. Failed to parse group id.");
                return 0;
            }
            if(!group_id) return 0;

            result.push_back(group_id);
            return 0;
        });
        LOG_SQL_CMD(res);
    }

    if(mode == GroupAssignmentCalculateMode::GLOBAL)
        if(auto parent = this->manager_->parent_manager(); parent) {
            auto parent_groups = parent->assignments().server_groups_of_client(mode, cldbid);
            result.reserve(result.size() + parent_groups.size());
            result.insert(result.begin(), parent_groups.begin(), parent_groups.end());
        }

    return result;
}

std::vector<ChannelGroupAssignment> GroupAssignmentManager::channel_groups_of_client(GroupAssignmentCalculateMode mode, ClientDbId cldbid) {
    std::vector<ChannelGroupAssignment> result{};
    bool cache_found{false};
    {
        std::lock_guard cache_lock{this->client_cache_lock};
        for(auto& entry : this->client_cache) {
            if(entry->client_database_id != cldbid) continue;

            result.reserve(entry->channel_group_assignments.size());
            result.insert(result.begin(), entry->channel_group_assignments.begin(), entry->channel_group_assignments.end());
            cache_found = true;
            break;
        }
    }

    if(!cache_found && !kCacheAllClients) {
        debugMessage(this->server_id(), "Query client groups for client {} on server {}.", cldbid, this->server_id());

        result.reserve(64);
        auto res = sql::command(this->sql_manager(), "SELECT `groupId`, `channelId`, `until` FROM `assignedGroups` WHERE `serverId` = :sid AND `cldbid` = :cldbid AND `channelId` != 0", variable{":sid", this->server_id()}, variable{":cldbid", cldbid})
                .query([&](int length, std::string* value, std::string* column) {
                    GroupId group_id{0};
                    ChannelId channel_id{0};
                    try {
                        for(int index = 0; index < length; index++) {
                            if(column[index] == "groupId") {
                                group_id = std::stoull(value[index]);
                            } else if(column[index] == "channelId") {
                                channel_id = std::stoull(value[index]);
                            }
                        }
                    } catch(std::exception& ex) {
                        logWarning(this->server_id(), "Invalid data found in group assignment table. Failed to parse group or channel id.");
                        return 0;
                    }
                    if(!group_id || !channel_id) return 0;

                    result.emplace_back(channel_id, group_id, false);
                    return 0;
                });
        LOG_SQL_CMD(res);
    }

    if(mode == GroupAssignmentCalculateMode::GLOBAL)
        if(auto parent = this->manager_->parent_manager(); parent) {
            auto parent_groups = parent->assignments().channel_groups_of_client(mode, cldbid);
            result.reserve(result.size() + parent_groups.size());
            result.insert(result.begin(), parent_groups.begin(), parent_groups.end());
        }

    return result;
}

std::deque<ts::ClientDbId> GroupAssignmentManager::server_group_clients(GroupId group_id) {
    std::deque<ts::ClientDbId> result{};
    if constexpr(kCacheAllClients) {
        std::lock_guard cache_lock{this->client_cache_lock};
        for(auto& client : this->client_cache) {
            auto it = std::find_if(client->server_group_assignments.begin(), client->server_group_assignments.end(), [&](const ServerGroupAssignment& assignment) {
                return assignment.group_id == group_id;
            });
            if(it == client->server_group_assignments.end())
                continue;
            result.push_back(client->client_database_id);
        }
    } else {
        auto res = sql::command(this->sql_manager(), "SELECT `cldbid` FROM `assignedGroups` WHERE `serverId` = :sid AND `channelId` = 0 AND `groupId` = :gid", variable{":sid", this->server_id()}, variable{":gid", group_id})
                .query([&](int length, std::string* value, std::string* column) {
                    ClientDbId cldbid{0};
                    try {
                        for(int index = 0; index < length; index++) {
                            if(column[index] == "cldbid") {
                                cldbid = std::stoull(value[index]);
                                break;
                            }
                        }
                    } catch(std::exception& ex) {
                        logWarning(this->server_id(), "Invalid data found in group assignment table. Failed to parse client database id.");
                        return 0;
                    }
                    if(!cldbid) return 0;

                    result.push_back(cldbid);
                    return 0;
                });
        LOG_SQL_CMD(res);
    }

    if(auto parent = this->manager_->parent_manager(); parent) {
        auto parent_clients = parent->assignments().server_group_clients(group_id);
        result.insert(result.begin(), parent_clients.begin(), parent_clients.end());
    }

    return result;
}

GroupAssignmentResult GroupAssignmentManager::add_server_group(ClientDbId client, GroupId group) {
    bool cache_verified{false};
    {
        std::lock_guard cache_lock{this->client_cache_lock};
        for(auto& entry : this->client_cache) {
            if(entry->client_database_id != client) continue;

            auto it = std::find_if(entry->server_group_assignments.begin(), entry->server_group_assignments.end(), [&](const ServerGroupAssignment& assignment) {
                return assignment.group_id == group;
            });
            if(it != entry->server_group_assignments.end())
                return GroupAssignmentResult::ADD_ALREADY_MEMBER_OF_GROUP;
            entry->server_group_assignments.emplace_back(group);
            cache_verified = true;
            break;
        }

        if(!cache_verified && kCacheAllClients) {
            /* add the client to the cache */
            auto cache = std::make_unique<ClientCache>();
            cache->client_database_id = client;
            cache->server_group_assignments.emplace_back(group);
            this->client_cache.push_back(std::move(cache));
        }
    }

    {

        auto command = sql::command(this->sql_manager(), "INSERT INTO `assignedGroups` (`serverId`, `cldbid`, `groupId`, `channelId`, `until`) VALUES (:sid, :cldbid, :gid, :chid, :until)",
                     variable{":sid", this->server_id()},
                     variable{":cldbid", client},
                     variable{":gid", group},
                     variable{":chid", 0},
                     variable{":until", 0});
        if(cache_verified)
            command.executeLater().waitAndGetLater(LOG_SQL_CMD, {-1, "failed to insert group assignment into the database"});
        else {
            auto result = command.execute();
            if(!result) return GroupAssignmentResult::ADD_ALREADY_MEMBER_OF_GROUP; //TODO: Parse error from database?
        }
    }
    return GroupAssignmentResult::SUCCESS;
}

GroupAssignmentResult GroupAssignmentManager::remove_server_group(ClientDbId client, GroupId group) {
    bool cache_verified{false};
    {
        std::lock_guard cache_lock{this->client_cache_lock};
        for(auto& entry : this->client_cache) {
            if(entry->client_database_id != client) continue;

            auto it = std::find_if(entry->server_group_assignments.begin(), entry->server_group_assignments.end(), [&](const ServerGroupAssignment& assignment) {
                return assignment.group_id == group;
            });
            if(it == entry->server_group_assignments.end())
                return GroupAssignmentResult::REMOVE_NOT_MEMBER_OF_GROUP;
            entry->server_group_assignments.erase(it);
            cache_verified = true;
            break;
        }

        if(!cache_verified && kCacheAllClients)
            return GroupAssignmentResult::REMOVE_NOT_MEMBER_OF_GROUP;
    }

    {
        auto command = sql::command(this->sql_manager(), "DELETE FROM `assignedGroups` WHERE `serverId` =  :sid AND `cldbid` = :cldbid AND `groupId` = :gid AND `channelId` = :chid",
                     variable{":sid", this->server_id()},
                     variable{":cldbid", client},
                     variable{":gid", group},
                     variable{":chid", 0});
        if(cache_verified)
            command.executeLater().waitAndGetLater(LOG_SQL_CMD, {-1, "failed to remove group assignment from database"});
        else {
            auto result = command.execute();
            if(!result) return GroupAssignmentResult::REMOVE_NOT_MEMBER_OF_GROUP; //TODO: Parse error from database?
        }
    }
    return GroupAssignmentResult::SUCCESS;
}

GroupAssignmentResult GroupAssignmentManager::set_channel_group(ClientDbId client, GroupId group, ChannelId channel_id, bool temporary) {
    bool cache_verified{false};
    {
        std::lock_guard cache_lock{this->client_cache_lock};
        for(auto& entry : this->client_cache) {
            if(entry->client_database_id != client) continue;

            auto it = std::find_if(entry->channel_group_assignments.begin(), entry->channel_group_assignments.end(), [&](const ChannelGroupAssignment& assignment) {
                return assignment.channel_id == channel_id;
            });
            if(it != entry->channel_group_assignments.end()) {
                if(group) {
                    if(it->group_id == group)
                        return GroupAssignmentResult::SET_ALREADY_MEMBER_OF_GROUP;
                    it->group_id = group;
                } else {
                    entry->channel_group_assignments.erase(it);
                }
            } else {
                if(group) {
                    entry->channel_group_assignments.emplace_back(channel_id, group, temporary);
                }
            }
            cache_verified = true;
            break;
        }

        if(!cache_verified && kCacheAllClients) {
            if(group) {
                /* add the client to the cache */
                auto cache = std::make_unique<ClientCache>();
                cache->client_database_id = client;
                cache->channel_group_assignments.emplace_back(channel_id, group, temporary);
                this->client_cache.push_back(std::move(cache));
            } else {
                return GroupAssignmentResult::SUCCESS;
            }
        }
    }
    if(temporary)
        return GroupAssignmentResult::SUCCESS;

    sql::command(this->sql_manager(), "DELETE FROM `assignedGroups` WHERE `serverId` =  :sid AND `cldbid` = :cldbid AND `channelId` = :chid",
                 variable{":sid", this->server_id()},
                 variable{":cldbid", client},
                 variable{":chid", channel_id})
            .executeLater().waitAndGetLater(LOG_SQL_CMD, {1, "failed to delete old channel group assignment"});

    if(group) {
        sql::command(this->sql_manager(), "INSERT INTO `assignedGroups` (`serverId`, `cldbid`, `groupId`, `channelId`, `until`) VALUES (:sid, :cldbid, :gid, :chid, :until)",
                     variable{":sid", this->server_id()},
                     variable{":cldbid", client},
                     variable{":gid", group},
                     variable{":chid", channel_id},
                     variable{":until", 0})
                .executeLater().waitAndGetLater(LOG_SQL_CMD, {1, "failed to insert channel group assignment"});
    }
    return GroupAssignmentResult::SUCCESS;
}

void GroupAssignmentManager::cleanup_channel_assignments(ChannelId channel_id) {
    {
        std::lock_guard cache_lock{this->client_cache_lock};
        for(auto& client : this->client_cache) {
            client->channel_group_assignments.erase(std::find_if(client->channel_group_assignments.begin(), client->channel_group_assignments.end(), [&](const ChannelGroupAssignment& assignment) {
                return assignment.channel_id == channel_id;
            }));
        }
    }

    sql::command(this->sql_manager(), "DELETE FROM `assignedGroups` WHERE `serverId` = :sid AND `channelId` = :cid", variable{":sid", this->server_id()}, variable{":cid", channel_id})
            .executeLater().waitAndGetLater(LOG_SQL_CMD, {1, "future failed"});
}

void GroupAssignmentManager::cleanup_assignments() {
    {

        std::lock_guard cache_lock{this->client_cache_lock};
        this->client_cache.clear();
    }

    sql::command(this->sql_manager(), "DELETE FROM `assignedGroups` WHERE `serverId` = :sid", variable{":sid", this->server_id()})
            .executeLater().waitAndGetLater(LOG_SQL_CMD, {1, "future failed"});
}

void GroupAssignmentManager::cleanup_channel_temporary_assignment(ClientDbId client_dbid, ChannelId channel) {
    std::lock_guard cache_lock{this->client_cache_lock};
    for(auto& client : this->client_cache) {
        if(client->client_database_id == client_dbid) {
            auto assignment = std::find_if(client->channel_group_assignments.begin(), client->channel_group_assignments.end(), [&](const ChannelGroupAssignment& assignment) {
                return assignment.channel_id == channel;
            });
            if(assignment->temporary_assignment)
                client->channel_group_assignments.erase(assignment);
            break;
        }
    }
}

std::deque<ts::property::ClientProperties> GroupAssignmentManager::update_client_group_properties(const std::shared_ptr<server::ConnectedClient> &client, ChannelId channel_id) {
    std::deque<property::ClientProperties> changed;
    bool increase_join_state{false};

    //Server groups
    {
        auto server_groups = this->server_groups_of_client(GroupAssignmentCalculateMode::GLOBAL, client->getClientDatabaseId());
        std::string group_string;
        for(const auto& group : server_groups)
            if(group_string.empty()) group_string += std::to_string(group);
            else group_string += "," + std::to_string(group);

        if(client->properties()[property::CLIENT_SERVERGROUPS] != group_string) {
            client->properties()[property::CLIENT_SERVERGROUPS] = group_string;
            changed.push_back(property::CLIENT_SERVERGROUPS);
            increase_join_state = true;
        }

    }

    //Channel group
    if(channel_id) {
        auto assignment = this->manager_->virtual_server()->channel_service().calculate_channel_group(channel_id, client->getClientDatabaseId(), client->getType());

        if(client->properties()[property::CLIENT_CHANNEL_GROUP_INHERITED_CHANNEL_ID] != assignment.inherited_channel_id) {
            client->properties()[property::CLIENT_CHANNEL_GROUP_INHERITED_CHANNEL_ID] = assignment.inherited_channel_id;
            changed.push_back(property::CLIENT_CHANNEL_GROUP_INHERITED_CHANNEL_ID);
        }

        if(client->properties()[property::CLIENT_CHANNEL_GROUP_ID] != assignment.group_id) {
            client->properties()[property::CLIENT_CHANNEL_GROUP_ID] = assignment.group_id;
            changed.push_back(property::CLIENT_CHANNEL_GROUP_ID);
            increase_join_state = true;
        }
    }

    if(increase_join_state)
        client->increase_join_state(); /* groups have changed :) */

    return changed;
}