//
// Created by WolverinDEV on 03/03/2020.
//
#include <log/LogUtils.h>
#include "GroupManager.h"
#include "../vserver/VirtualServerBase.h"
#include "../InstanceHandler.h"

using namespace ts::server::groups;

GroupManager::GroupManager(ts::server::vserver::VirtualServerBase *server, std::shared_ptr<GroupManager> parent)
    : virtual_server_{server}, parent_manager_{std::move(parent)}, assignment_manager_{this} {
    assert(server);
}

GroupManager::~GroupManager() = default;

ts::ServerId GroupManager::server_id() {
    return this->virtual_server_->server_id();
}

sql::SqlManager* GroupManager::sql_manager() {
    return this->virtual_server_->sql_manager();
}

bool GroupManager::initialize(const std::shared_ptr<GroupManager>& self, std::string &error) {
    assert(&*self == this);
    this->weak_ref_ = self;
    return this->assignment_manager_.initialize(error);
}

GroupLoadResult GroupManager::load_data(bool initialize) {
    std::lock_guard manage_lock{this->group_manage_lock_};
    {
        std::lock_guard list_lock{this->group_lock_};
        this->server_groups_.clear();
        this->channel_groups_.clear();
    }

    auto res = sql::command(this->sql_manager(), "SELECT * FROM `groups` WHERE `serverId` = :sid", variable{":sid", this->server_id()}).query(&GroupManager::insert_group_from_sql, this);
    if(!res) {
        LOG_SQL_CMD(res);
        return GroupLoadResult::DATABASE_ERROR;
    }

    if(this->server_groups_.empty() || this->channel_groups_.empty())
        return GroupLoadResult::NO_GROUPS;

    this->ensure_valid_default_groups(initialize);
    return GroupLoadResult::SUCCESS;
}

void GroupManager::unload_data() {
    std::lock_guard manage_lock{this->group_manage_lock_};
    {
        std::lock_guard list_lock{this->group_lock_};
        this->server_groups_.clear();
        this->channel_groups_.clear();
    }
}

void GroupManager::reset_groups(bool db_cleanup) {
    std::lock_guard manage_lock{this->group_manage_lock_};
    this->unload_data();

    if(db_cleanup) {
        LOG_SQL_CMD(sql::command(this->virtual_server_->sql_manager(), "DELETE FROM `permissions` WHERE `serverId` = :serverId AND `type` = :type",
                                 variable{":serverId", this->server_id()},
                                 variable{":type", ts::permission::SQL_PERM_GROUP}).execute());
        LOG_SQL_CMD(sql::command(this->virtual_server_->sql_manager(), "DELETE FROM `assignedGroups` WHERE `serverId` = :serverId", variable{":serverId", this->server_id()}).execute());
        LOG_SQL_CMD(sql::command(this->virtual_server_->sql_manager(), "DELETE FROM `groups` WHERE `serverId` = :serverId", variable{":serverId", this->server_id()}).execute());
    }

    if(auto error = this->load_data(true); error != GroupLoadResult::SUCCESS)
        logCritical(this->server_id(), "Failed to load groups after group unload ({}). There might be no groups loaded now!", (int) error);
}

int GroupManager::insert_group_from_sql(int length, std::string *values, std::string *names) {
    GroupId group_id{0};
    GroupType group_type{GroupType::GROUP_TYPE_UNKNOWN};
    std::string group_name{};
    GroupTarget group_target{GroupTarget::GROUPTARGET_UNKNOWN};

    for(size_t index = 0; index < length; index++) {
        try {
            if(names[index] == "groupId")
                group_id = std::stoull(values[index]);
            else if(names[index] == "target")
                group_target = (GroupTarget) std::stoull(values[index]);
            else if(names[index] == "type")
                group_type = (GroupType) std::stoull(values[index]);
            else if(names[index] == "displayName")
                group_name = names[index];
        } catch(std::exception& ex) {
            logWarning(this->server_id(), "Failed to parse group from database. Failed to parse column {} (value: {})", names[index], values[index]);
            return 0;
        }
    }

    if(!group_id || group_type == GroupType::GROUP_TYPE_UNKNOWN || group_target == GroupTarget::GROUPTARGET_UNKNOWN) {
        logWarning(this->server_id(), "Failed to query group from database. Invalid values found.");
        return 0;
    }

    auto database_target = (uint8_t) (group_target == GroupTarget::GROUPTARGET_SERVER ? DatabaseGroupTarget::SERVER : DatabaseGroupTarget::CHANNEL);
    auto permissions = serverInstance->databaseHelper()->loadGroupPermissions(this->server_id(), group_id, database_target);

    if(group_target == GroupTarget::GROUPTARGET_SERVER) {
        auto group = std::make_shared<ServerGroup>(this->server_id(), group_id, group_type, group_name, permissions);

        std::lock_guard lock{this->group_lock_};
        this->server_groups_.push_back(group);
    } else {
        auto group = std::make_shared<ChannelGroup>(this->server_id(), group_id, group_type, group_name, permissions);

        std::lock_guard lock{this->group_lock_};
        this->channel_groups_.push_back(group);
    }

    return 0;
}

void GroupManager::ensure_valid_default_groups(bool initialize) {
    assert(!this->channel_groups_.empty() && !this->server_groups_.empty());

    auto fallback_server_group = this->server_groups_.front();
    auto fallback_channel_group = this->channel_groups_.front();

    auto properties = this->virtual_server_->server_properties();
    /* default server music group */
    {
        auto group_id = properties->find(property::VIRTUALSERVER_DEFAULT_MUSIC_GROUP).as_save<GroupId>();
        auto group = this->find_server_group(groups::GroupCalculateMode::GLOBAL, group_id);
        if(!group) {
            if(!initialize)
                logWarning(this->server_id(), "Missing default music bot group ({}). Try to reinitialize from serverinstance_template_musicdefault_group.", group_id);

            group_id = serverInstance->properties()[property::SERVERINSTANCE_TEMPLATE_MUSICDEFAULT_GROUP].as<GroupId>();
            group = this->find_server_group(groups::GroupCalculateMode::GLOBAL, group_id);
            if(!group) {
                group = fallback_server_group;
                logCritical(LOG_INSTANCE, "Failed to find default music bot group ({}). Using first available group ({}).", group_id, fallback_server_group->group_id());
            }

            auto group_name = group->display_name();
            group = this->find_server_group_by_name(GroupCalculateMode::LOCAL, group_name);
            if(!group) {
                group = fallback_server_group;
                logError(this->server_id(), "Missing default music bot group and could not find group from template name ({}). Using {} as fallback.", group_name, group);
            }

            properties->find(property::VIRTUALSERVER_DEFAULT_MUSIC_GROUP) = group->group_id();
        }
    }

    /* default server guest group */
    {
        auto group_id = properties->find(property::VIRTUALSERVER_DEFAULT_SERVER_GROUP).as_save<GroupId>();
        auto group = this->find_server_group(groups::GroupCalculateMode::GLOBAL, group_id);
        if(!group) {
            if(!initialize)
                logWarning(this->server_id(), "Missing default server guest group ({}). Try to reinitialize from serverinstance_template_serverdefault_group.", group_id);

            group_id = serverInstance->properties()[property::SERVERINSTANCE_TEMPLATE_SERVERDEFAULT_GROUP].as<GroupId>();
            group = this->find_server_group(groups::GroupCalculateMode::GLOBAL, group_id);
            if(!group) {
                group = fallback_server_group;
                logCritical(LOG_INSTANCE, "Failed to find default server guest group ({}). Using first available group ({}).", group_id, fallback_server_group->group_id());
            }

            auto group_name = group->display_name();
            group = this->find_server_group_by_name(GroupCalculateMode::LOCAL, group_name);
            if(!group) {
                group = fallback_server_group;
                logError(this->server_id(), "Missing default server guest group and could not find group from template name ({}). Using {} as fallback.", group_name, group);
            }

            properties->find(property::VIRTUALSERVER_DEFAULT_SERVER_GROUP) = group->group_id();
        }
    }

    /* default channel admin group */
    {
        auto group_id = properties->find(property::VIRTUALSERVER_DEFAULT_CHANNEL_ADMIN_GROUP).as_save<GroupId>();
        auto group = this->find_channel_group(groups::GroupCalculateMode::GLOBAL, group_id);
        if(!group) {
            if(!initialize)
                logWarning(this->server_id(), "Missing default channel admin group ({}). Try to reinitialize from serverinstance_template_channeladmin_group.", group_id);

            group_id = serverInstance->properties()[property::SERVERINSTANCE_TEMPLATE_CHANNELADMIN_GROUP].as<GroupId>();
            group = this->find_channel_group(groups::GroupCalculateMode::GLOBAL, group_id);
            if(!group) {
                group = fallback_channel_group;
                logCritical(LOG_INSTANCE, "Failed to find default channel admin group ({}). Using first available group ({}).", group_id, fallback_channel_group->group_id());
            }

            auto group_name = group->display_name();
            group = this->find_channel_group_by_name(GroupCalculateMode::LOCAL, group_name);
            if(!group) {
                group = fallback_channel_group;
                logError(this->server_id(), "Missing default channel admin group and could not find group from template name ({}). Using {} as fallback.", group_name, group);
            }

            properties->find(property::VIRTUALSERVER_DEFAULT_CHANNEL_ADMIN_GROUP) = group->group_id();
        }
    }

    /* default channel group */
    {
        auto group_id = properties->find(property::VIRTUALSERVER_DEFAULT_CHANNEL_GROUP).as_save<GroupId>();
        auto group = this->find_channel_group(groups::GroupCalculateMode::GLOBAL, group_id);
        if(!group) {
            if(!initialize)
                logWarning(this->server_id(), "Missing default channel guest group ({}). Try to reinitialize from serverinstance_template_channeldefault_group.", group_id);

            group_id = serverInstance->properties()[property::SERVERINSTANCE_TEMPLATE_CHANNELDEFAULT_GROUP].as<GroupId>();
            group = this->find_channel_group(groups::GroupCalculateMode::GLOBAL, group_id);
            if(!group) {
                group = fallback_channel_group;
                logCritical(LOG_INSTANCE, "Failed to find default channel guest group ({}). Using first available group ({}).", group_id, fallback_channel_group->group_id());
            }

            auto group_name = group->display_name();
            group = this->find_channel_group_by_name(GroupCalculateMode::LOCAL, group_name);
            if(!group) {
                group = fallback_channel_group;
                logError(this->server_id(), "Missing default channel guest group and could not find group from template name ({}). Using {} as fallback.", group_name, group);
            }

            properties->find(property::VIRTUALSERVER_DEFAULT_CHANNEL_GROUP) = group->group_id();
        }
    }
}

std::shared_ptr<ServerGroup> GroupManager::default_server_group(ClientType client_type, bool enforce_id) {
    auto id =
            client_type != ClientType::CLIENT_QUERY ?
            this->virtual_server_->server_properties()->find(property::VIRTUALSERVER_DEFAULT_SERVER_GROUP).as_save<GroupId>() :
            serverInstance->properties()[property::SERVERINSTANCE_GUEST_SERVERQUERY_GROUP].as_save<GroupId>();
    auto group = this->find_server_group(groups::GroupCalculateMode::GLOBAL, id);
    if(group || enforce_id) return group;

    std::lock_guard glock{this->group_lock_};
    return this->server_groups_.empty() ? nullptr : this->server_groups_.front();
}

std::shared_ptr<ChannelGroup> GroupManager::default_channel_group(ClientType client_type, bool enforce_id) {
    auto id = this->virtual_server_->server_properties()->find(property::VIRTUALSERVER_DEFAULT_CHANNEL_GROUP).as_save<GroupId>();
    auto group = this->find_channel_group(groups::GroupCalculateMode::GLOBAL, id);
    if(group || enforce_id) return group;

    std::lock_guard glock{this->group_lock_};
    return this->channel_groups_.empty() ? nullptr : this->channel_groups_.front();
}

std::shared_ptr<ServerGroup> GroupManager::find_server_group(GroupCalculateMode mode, GroupId group_id) {
    {
        std::lock_guard glock{this->group_lock_};
        auto it = std::find_if(this->server_groups_.begin(), this->server_groups_.end(), [&](const std::shared_ptr<ServerGroup>& group) {
            return group->group_id() == group_id;
        });

        if(it != this->server_groups_.end())
            return *it;
    }
    if(mode == GroupCalculateMode::GLOBAL && this->parent_manager_)
        return this->parent_manager_->find_server_group(mode, group_id);
    return nullptr;
}

std::shared_ptr<ServerGroup> GroupManager::find_server_group_by_name(GroupCalculateMode mode, const std::string &name) {
    {
        std::string lname{name};
        std::transform(lname.begin(), lname.end(), lname.begin(), ::tolower);

        std::lock_guard glock{this->group_lock_};
        auto it = std::find_if(this->server_groups_.begin(), this->server_groups_.end(),
                               [&](const std::shared_ptr<ServerGroup> &group) {
                                   std::string lgroup_name{group->name_};
                                   std::transform(lgroup_name.begin(), lgroup_name.end(), lgroup_name.begin(),
                                                  ::tolower);
                                   return lname == lgroup_name;
                               });

        if(it != this->server_groups_.end())
            return *it;
    }

    if(mode == GroupCalculateMode::GLOBAL && this->parent_manager_)
        return this->parent_manager_->find_server_group_by_name(mode, name);
    return nullptr;
}

std::shared_ptr<ChannelGroup> GroupManager::find_channel_group(GroupCalculateMode mode, GroupId group_id) {
    {
        std::lock_guard glock{this->group_lock_};
        auto it = std::find_if(this->channel_groups_.begin(), this->channel_groups_.end(), [&](const std::shared_ptr<ChannelGroup>& group) {
            return group->group_id() == group_id;
        });

        if(it != this->channel_groups_.end())
            return *it;
    }

    if(mode == GroupCalculateMode::GLOBAL && this->parent_manager_)
        return this->parent_manager_->find_channel_group(mode, group_id);
    return nullptr;
}

std::shared_ptr<ChannelGroup> GroupManager::find_channel_group_by_name(GroupCalculateMode mode, const std::string &name) {
    {
        std::string lname{name};
        std::transform(lname.begin(), lname.end(), lname.begin(), ::tolower);

        std::lock_guard glock{this->group_lock_};
        auto it = std::find_if(this->channel_groups_.begin(), this->channel_groups_.end(),
                               [&](const std::shared_ptr<ChannelGroup> &group) {
                                   std::string lgroup_name{group->name_};
                                   std::transform(lgroup_name.begin(), lgroup_name.end(), lgroup_name.begin(),
                                                  ::tolower);
                                   return lname == lgroup_name;
                               });

        if(it != this->channel_groups_.end())
            return *it;
    }

    if(mode == GroupCalculateMode::GLOBAL && this->parent_manager_)
        return this->parent_manager_->find_channel_group_by_name(mode, name);
    return nullptr;
}

GroupCreateResult GroupManager::create_server_group(GroupType type, const std::string &name, std::shared_ptr<ServerGroup>& result) {
    std::lock_guard manage_lock{this->group_manage_lock_};
    auto self = this->weak_ref_.lock();
    if(!self) return GroupCreateResult::DATABASE_ERROR;

    if(this->find_server_group_by_name(GroupCalculateMode::LOCAL, name))
        return GroupCreateResult::NAME_ALREADY_IN_USED;

    auto group_id = this->generate_server_group_id();
    if(!group_id)
        return GroupCreateResult::FAILED_TO_GENERATE_ID;

    auto res = sql::command(this->sql_manager(), "INSERT INTO `groups` (`serverId`, `groupId`, `target`, `type`, `displayName`) VALUES (:sid, :gid, :target, :type, :name)",
                            variable{":sid", this->server_id()}, variable{":gid", group_id}, variable{":target", DatabaseGroupTarget::SERVER}, variable{":type", type}, variable{":name", name}).execute();

    if(!res) {
        LOG_SQL_CMD(res);
        return GroupCreateResult::DATABASE_ERROR;
    }

    auto permissions = serverInstance->databaseHelper()->loadGroupPermissions(this->server_id(), group_id, (uint8_t) DatabaseGroupTarget::SERVER);
    auto group = std::make_shared<ServerGroup>(self->server_id(), group_id, type, name, permissions);
    {
        std::lock_guard glock{this->group_lock_};
        this->server_groups_.push_back(group);
    }
    result = group;
    return GroupCreateResult::SUCCESS;
}

GroupCreateResult GroupManager::create_channel_group(GroupType type, const std::string &name, std::shared_ptr<ChannelGroup> &result) {
    std::lock_guard manage_lock{this->group_manage_lock_};
    auto self = this->weak_ref_.lock();
    if(!self) return GroupCreateResult::DATABASE_ERROR;

    if(this->find_channel_group_by_name(GroupCalculateMode::LOCAL, name))
        return GroupCreateResult::NAME_ALREADY_IN_USED;

    auto group_id = this->generate_channel_group_id();
    if(!group_id)
        return GroupCreateResult::FAILED_TO_GENERATE_ID;

    auto res = sql::command(this->sql_manager(), "INSERT INTO `groups` (`serverId`, `groupId`, `target`, `type`, `displayName`) VALUES (:sid, :gid, :target, :type, :name)",
                            variable{":sid", this->server_id()}, variable{":gid", group_id}, variable{":target", DatabaseGroupTarget::CHANNEL}, variable{":type", type}, variable{":name", name}).execute();

    if(!res) {
        LOG_SQL_CMD(res);
        return GroupCreateResult::DATABASE_ERROR;
    }

    auto permissions = serverInstance->databaseHelper()->loadGroupPermissions(this->server_id(), group_id, (uint8_t) DatabaseGroupTarget::CHANNEL);
    auto group = std::make_shared<ChannelGroup>(this->server_id(), group_id, type, name, permissions);
    {
        std::lock_guard glock{this->group_lock_};
        this->channel_groups_.push_back(group);
    }
    result = group;
    return GroupCreateResult::SUCCESS;
}

GroupCopyResult GroupManager::copy_server_group(GroupId source, GroupType target_type, const std::string &display_name, std::shared_ptr<ServerGroup>& result) {
    std::lock_guard manage_lock{this->group_manage_lock_};
    auto self = this->weak_ref_.lock();
    if(!self) return GroupCopyResult::DATABASE_ERROR;

    std::shared_ptr<ServerGroup> group{};
    auto create_result = this->create_server_group(target_type, display_name, group);
    switch(create_result) {
        case GroupCreateResult::SUCCESS:
            break;

        case GroupCreateResult::DATABASE_ERROR:
            return GroupCopyResult::DATABASE_ERROR;

        case GroupCreateResult::NAME_ALREADY_IN_USED:
            return GroupCopyResult::NAME_ALREADY_IN_USE;

        case GroupCreateResult::FAILED_TO_GENERATE_ID:
            return GroupCopyResult::FAILED_TO_GENERATE_ID;
    }

    assert(group);
    return this->copy_server_group_permissions(source, group->group_id());
}

GroupCopyResult GroupManager::copy_channel_group(GroupId source, GroupType target_type, const std::string &display_name, std::shared_ptr<ChannelGroup> &result) {
    std::lock_guard manage_lock{this->group_manage_lock_};
    auto self = this->weak_ref_.lock();
    if(!self) return GroupCopyResult::DATABASE_ERROR;

    std::shared_ptr<ChannelGroup> group{};
    auto create_result = this->create_channel_group(target_type, display_name, group);
    switch(create_result) {
        case GroupCreateResult::SUCCESS:
            break;

        case GroupCreateResult::DATABASE_ERROR:
            return GroupCopyResult::DATABASE_ERROR;

        case GroupCreateResult::NAME_ALREADY_IN_USED:
            return GroupCopyResult::NAME_ALREADY_IN_USE;

        case GroupCreateResult::FAILED_TO_GENERATE_ID:
            return GroupCopyResult::FAILED_TO_GENERATE_ID;
    }

    assert(group);
    return this->copy_channel_group_permissions(source, group->group_id());
}

GroupCopyResult GroupManager::copy_server_group_permissions(GroupId source, GroupId target) {
    std::lock_guard manage_lock{this->group_manage_lock_};
    auto self = this->weak_ref_.lock();
    if(!self) return GroupCopyResult::DATABASE_ERROR;

    auto source_group = this->find_server_group(groups::GroupCalculateMode::GLOBAL, source);
    if(!source_group)
        return GroupCopyResult::UNKNOWN_SOURCE_GROUP;

    auto target_group = this->find_server_group(groups::GroupCalculateMode::GLOBAL, target);
    if(!target_group)
        return GroupCopyResult::UNKNOWN_TARGET_GROUP;

    auto res = sql::command(this->sql_manager(), "DELETE FROM `permissions` WHERE `serverId` = :sid AND `type` = :type AND `id` = :id",
            variable{":sid", this->server_id()}, variable{":type", ts::permission::SQL_PERM_GROUP}, variable{":id", target}).execute();
    if(!res) {
        LOG_SQL_CMD(res);
        return GroupCopyResult::DATABASE_ERROR;
    }

    res = sql::command(this->sql_manager(), "INSERT INTO `permissions` (`serverId`, `type`, `id`, `channelId`, `permId`, `value`, `grant`, `flag_skip`, `flag_negate`) SELECT :tsid AS `serverId`, `type`, :target AS `id`, 0 AS `channelId`, `permId`, `value`,`grant`,`flag_skip`, `flag_negate` FROM `permissions` WHERE `serverId` = :ssid AND `type` = :type AND `id` = :source",
                       variable{":ssid", source_group->virtual_server_id()},
                       variable{":tsid", target_group->virtual_server_id()},
                       variable{":type", ts::permission::SQL_PERM_GROUP},
                       variable{":source", source},
                       variable{":target", target}).execute();
    target_group->set_permissions(serverInstance->databaseHelper()->loadGroupPermissions(this->server_id(), target, (uint8_t) DatabaseGroupTarget::SERVER));
    LOG_SQL_CMD(res);

    return GroupCopyResult::SUCCESS;
}

GroupCopyResult GroupManager::copy_channel_group_permissions(GroupId source, GroupId target) {
    std::lock_guard manage_lock{this->group_manage_lock_};
    auto self = this->weak_ref_.lock();
    if(!self) return GroupCopyResult::DATABASE_ERROR;

    auto source_group = this->find_channel_group(groups::GroupCalculateMode::GLOBAL, source);
    if(!source_group)
        return GroupCopyResult::UNKNOWN_SOURCE_GROUP;

    auto target_group = this->find_channel_group(groups::GroupCalculateMode::GLOBAL, target);
    if(!target_group)
        return GroupCopyResult::UNKNOWN_TARGET_GROUP;

    auto res = sql::command(this->sql_manager(), "DELETE FROM `permissions` WHERE `serverId` = :sid AND `type` = :type AND `id` = :id",
                            variable{":sid", this->server_id()}, variable{":type", ts::permission::SQL_PERM_GROUP}, variable{":id", target}).execute();
    if(!res) {
        LOG_SQL_CMD(res);
        return GroupCopyResult::DATABASE_ERROR;
    }

    res = sql::command(this->sql_manager(), "INSERT INTO `permissions` (`serverId`, `type`, `id`, `channelId`, `permId`, `value`, `grant`, `flag_skip`, `flag_negate`) SELECT :tsid AS `serverId`, `type`, :target AS `id`, 0 AS `channelId`, `permId`, `value`,`grant`,`flag_skip`, `flag_negate` FROM `permissions` WHERE `serverId` = :ssid AND `type` = :type AND `id` = :source",
                       variable{":ssid", source_group->virtual_server_id()},
                       variable{":tsid", target_group->virtual_server_id()},
                       variable{":type", ts::permission::SQL_PERM_GROUP},
                       variable{":source", source},
                       variable{":target", target}).execute();
    target_group->set_permissions(serverInstance->databaseHelper()->loadGroupPermissions(this->server_id(), target, (uint8_t) DatabaseGroupTarget::CHANNEL));
    LOG_SQL_CMD(res);

    return GroupCopyResult::SUCCESS;
}


GroupRenameResult GroupManager::rename_server_group(GroupId group_id, const std::string &name) {
    std::lock_guard manage_lock{this->group_manage_lock_};
    auto self = this->weak_ref_.lock();
    if(!self) return GroupRenameResult::DATABASE_ERROR;

    if(name.empty() || name.length() > 40)
        return GroupRenameResult::NAME_INVALID;

    auto group = this->find_server_group(groups::GroupCalculateMode::GLOBAL, group_id);
    if(!group) return GroupRenameResult::INVALID_GROUP_ID;

    if(this->find_server_group_by_name(GroupCalculateMode::LOCAL, name))
        return GroupRenameResult::NAME_ALREADY_USED;

    sql::command(this->sql_manager(), "UPDATE `groups` SET `displayName` = :name WHERE `serverId` = :server AND `groupId` = :group_id AND `target` = :target",
            variable{":server", this->server_id()},
            variable{":target", DatabaseGroupTarget::SERVER},
            variable{":group_id", group_id}).executeLater().waitAndGetLater(LOG_SQL_CMD, {-1, "future failed"});
    group->name_ = name;
    return GroupRenameResult::SUCCESS;
}

GroupRenameResult GroupManager::rename_channel_group(GroupId group_id, const std::string &name) {
    std::lock_guard manage_lock{this->group_manage_lock_};
    auto self = this->weak_ref_.lock();
    if(!self) return GroupRenameResult::DATABASE_ERROR;

    if(name.empty() || name.length() > 40)
        return GroupRenameResult::NAME_INVALID;

    auto group = this->find_channel_group(groups::GroupCalculateMode::GLOBAL, group_id);
    if(!group) return GroupRenameResult::INVALID_GROUP_ID;

    if(this->find_channel_group_by_name(GroupCalculateMode::LOCAL, name))
        return GroupRenameResult::NAME_ALREADY_USED;

    sql::command(this->sql_manager(), "UPDATE `groups` SET `displayName` = :name WHERE `serverId` = :server AND `groupId` = :group_id AND `target` = :target",
                 variable{":server", this->server_id()},
                 variable{":target", DatabaseGroupTarget::CHANNEL},
                 variable{":group_id", group_id}).executeLater().waitAndGetLater(LOG_SQL_CMD, {-1, "future failed"});
    group->name_ = name;
    return GroupRenameResult::SUCCESS;
}

GroupDeleteResult GroupManager::delete_server_group(GroupId group_id) {
    std::lock_guard manage_lock{this->group_manage_lock_};
    auto self = this->weak_ref_.lock();
    if(!self) return GroupDeleteResult::DATABASE_ERROR;

    {
        std::lock_guard glock{this->group_lock_};
        auto it = std::find_if(this->server_groups_.begin(), this->server_groups_.begin(), [&](const std::shared_ptr<ServerGroup>& group) {
            return group->group_id() == group_id;
        });
        if(it == this->server_groups_.end())
            return GroupDeleteResult::INVALID_GROUP_ID;
        this->server_groups_.erase(it);
    }

    sql::command(this->sql_manager(), "DELETE FROM WHERE `serverId` = :server AND `groupId` = :group_id AND `target` = :target",
                 variable{":server", this->server_id()},
                 variable{":target", DatabaseGroupTarget::SERVER},
                 variable{":group_id", group_id}).executeLater().waitAndGetLater(LOG_SQL_CMD, {-1, "future failed"});
    return GroupDeleteResult::SUCCESS;
}

GroupDeleteResult GroupManager::delete_channel_group(GroupId group_id) {
    std::lock_guard manage_lock{this->group_manage_lock_};
    auto self = this->weak_ref_.lock();
    if(!self) return GroupDeleteResult::DATABASE_ERROR;

    {
        std::lock_guard glock{this->group_lock_};
        auto it = std::find_if(this->channel_groups_.begin(), this->channel_groups_.begin(), [&](const std::shared_ptr<ChannelGroup>& group) {
            return group->group_id() == group_id;
        });
        if(it == this->channel_groups_.end())
            return GroupDeleteResult::INVALID_GROUP_ID;
        this->channel_groups_.erase(it);
    }

    sql::command(this->sql_manager(), "DELETE FROM WHERE `serverId` = :server AND `groupId` = :group_id AND `target` = :target",
                 variable{":server", this->server_id()},
                 variable{":target", DatabaseGroupTarget::CHANNEL},
                 variable{":group_id", group_id}).executeLater().waitAndGetLater(LOG_SQL_CMD, {-1, "future failed"});
    return GroupDeleteResult::SUCCESS;
}

ts::GroupId GroupManager::generate_server_group_id() {
    std::lock_guard glock{this->group_lock_};
    ts::GroupId highest_group_id{0};
    for(auto& group : this->channel_groups_)
        highest_group_id = std::max(group->group_id(), highest_group_id);

    for(auto& group : this->server_groups_)
        highest_group_id = std::max(group->group_id(), highest_group_id);

    if(this->parent_manager_)
        highest_group_id = std::max(this->parent_manager_->generate_server_group_id(), highest_group_id);
    return highest_group_id + 1;
}

ts::GroupId GroupManager::generate_channel_group_id() {
    return this->generate_server_group_id();
}