#pragma once

#include <mutex>
#include <deque>
#include <string>
#include <memory>
#include <Definitions.h>
#include <sql/SqlQuery.h>
#include "./GroupAssignmentManager.h"
#include "./Group.h"

namespace ts::server {
    namespace vserver {
        class VirtualServerBase;
    }

    namespace groups {
        enum struct GroupCalculateMode {
            LOCAL, /* only calculate clients groups for the local server */
            GLOBAL /* use the parent group manager as well, if existing */
        };

        enum struct GroupLoadResult {
            SUCCESS,
            NO_GROUPS,
            DATABASE_ERROR
        };

        enum struct GroupCreateResult {
            SUCCESS,
            NAME_ALREADY_IN_USED,
            FAILED_TO_GENERATE_ID,
            DATABASE_ERROR
        };

        enum struct GroupCopyResult {
            SUCCESS,
            UNKNOWN_SOURCE_GROUP,
            UNKNOWN_TARGET_GROUP,
            NAME_ALREADY_IN_USE,
            FAILED_TO_GENERATE_ID,
            DATABASE_ERROR
        };

        enum struct GroupRenameResult {
            SUCCESS,
            INVALID_GROUP_ID,
            NAME_ALREADY_USED,
            NAME_INVALID,
            DATABASE_ERROR
        };

        enum struct GroupDeleteResult {
            SUCCESS,
            INVALID_GROUP_ID,
            DATABASE_ERROR
        };

        class GroupManager {
                friend class Group;
            public:
                enum struct DatabaseGroupTarget {
                    SERVER,
                    CHANNEL
                };

                GroupManager(vserver::VirtualServerBase* /* virtual server */, std::shared_ptr<GroupManager> /* parent */);
                ~GroupManager();

                bool initialize(const std::shared_ptr<GroupManager>& /* self ref, */, std::string& /* error */);
                GroupLoadResult load_data(bool /* initialize */ = false);
                void unload_data();
                void reset_groups(bool /* cleanup database */);

                [[nodiscard]] inline vserver::VirtualServerBase* virtual_server() { return this->virtual_server_; }

                [[nodiscard]] inline const std::shared_ptr<GroupManager>& parent_manager() { return this->parent_manager_; }
                [[nodiscard]] inline GroupAssignmentManager& assignments() { return this->assignment_manager_; }

                [[nodiscard]] std::vector<std::shared_ptr<ServerGroup>> server_groups(GroupCalculateMode /* mode */);
                [[nodiscard]] std::vector<std::shared_ptr<ChannelGroup>> channel_groups(GroupCalculateMode /* mode */);

                [[nodiscard]] std::shared_ptr<ServerGroup> default_server_group(ClientType /* client type */, bool /* enforce id */ = false);
                [[nodiscard]] std::shared_ptr<ChannelGroup> default_channel_group(ClientType /* client type */, bool /* enforce id */ = false);

                [[nodiscard]] std::shared_ptr<ServerGroup> find_server_group(GroupCalculateMode /* mode */, GroupId /* group id */);
                [[nodiscard]] std::shared_ptr<ChannelGroup> find_channel_group(GroupCalculateMode /* mode */, GroupId /* group id */);

                [[nodiscard]] std::shared_ptr<ServerGroup> find_server_group_by_name(GroupCalculateMode /* mode */, const std::string& /* group name */);
                [[nodiscard]] std::shared_ptr<ChannelGroup> find_channel_group_by_name(GroupCalculateMode /* mode */, const std::string& /* group name */);

                [[nodiscard]] GroupCreateResult create_server_group(GroupType type, const std::string& /* group name */, std::shared_ptr<ServerGroup>& /* result */);
                [[nodiscard]] GroupCreateResult create_channel_group(GroupType type, const std::string& /* group name */, std::shared_ptr<ChannelGroup>& /* result */);

                [[nodiscard]] GroupCopyResult copy_server_group(GroupId /* group id */, GroupType /* target group type */, const std::string& /* target group name */, std::shared_ptr<ServerGroup>& /* result */);
                [[nodiscard]] GroupCopyResult copy_channel_group(GroupId /* group id */, GroupType /* target group type */, const std::string& /* target group name */, std::shared_ptr<ChannelGroup>& /* result */);

                [[nodiscard]] GroupCopyResult copy_server_group_permissions(GroupId /* group id */, GroupId /* target group */);
                [[nodiscard]] GroupCopyResult copy_channel_group_permissions(GroupId /* group id */, GroupId /* target group */);

                [[nodiscard]] GroupRenameResult rename_server_group(GroupId /* group id */, const std::string& /* target group name */);
                [[nodiscard]] GroupRenameResult rename_channel_group(GroupId /* group id */, const std::string& /* target group name */);

                [[nodiscard]] GroupDeleteResult delete_server_group(GroupId /* group id */);
                [[nodiscard]] GroupDeleteResult delete_channel_group(GroupId /* group id */);
            private:
                vserver::VirtualServerBase* virtual_server_;

                std::weak_ptr<GroupManager> weak_ref_;
                std::shared_ptr<GroupManager> parent_manager_;
                GroupAssignmentManager assignment_manager_;

                /* recursive_mutex due to the copy server/channel group methods */
                std::recursive_mutex group_manage_lock_{};
                std::mutex group_lock_{};

                /* I think std::vector is better here because we will iterate more often than add groups */
                std::vector<std::shared_ptr<ServerGroup>> server_groups_{};
                std::vector<std::shared_ptr<ChannelGroup>> channel_groups_{};

                [[nodiscard]] sql::SqlManager* sql_manager();
                [[nodiscard]] ServerId server_id();

                void ensure_valid_default_groups(bool /* initialize */);

                int insert_group_from_sql(int /* length */, std::string* /* values */, std::string* /* columns */);
                [[nodiscard]] GroupId generate_server_group_id();
                [[nodiscard]] GroupId generate_channel_group_id();
        };
    }
}

DEFINE_TRANSFORMS(ts::server::groups::GroupManager::DatabaseGroupTarget, uint8_t);