#pragma once

#include <thread>
#include <event.h>
#include <vector>
#include <misc/spin_lock.h>
#include <Definitions.h>
#include <mutex>
#include <deque>

namespace ts::server {
    class VirtualServer;
}

namespace ts::connection {
    class VoiceClientConnection;
}

namespace ts::server::vserver {
    class VirtualServerBase;
}

namespace ts::server::server::udp {
    struct datagram_packet {
        union pktinfo_storage {
            in_pktinfo v4;
            in6_pktinfo v6;
        };

        datagram_packet* next_packet;

        sockaddr_storage address;
        pktinfo_storage address_info;

        size_t data_length;
        uint8_t data[0];
    };
    static_assert(std::is_trivially_destructible<datagram_packet>::value);
    static_assert(std::is_trivially_constructible<datagram_packet>::value);

    template <typename T, size_t N>
    struct write_ring_queue {
        std::array<T, N> memory{};
        size_t current_index{0};
        size_t filled_index{0};

        [[nodiscard]] constexpr inline auto max_size() const { return N; }
        [[nodiscard]] inline size_t current_size() const { return this->filled_index - this->current_index; }

        [[nodiscard]] inline bool pop_entry(T& result) {
            if(this->current_index >= this->filled_index) return false;
            return this->memory[this->current_index++ % N];
        }

        [[nodiscard]] inline bool push_entry(T&& entry) {
            if(this->filled_index - this->current_index >= N) return false;
            this->memory[this->filled_index++ % N] = std::forward(entry);
            return true;
        }
    };

    struct io_loop;
    struct io_loop_entry {
        io_loop* io_loop{nullptr};

        int file_descriptor{0};

        struct event* event_read{};
        struct event* event_write{};

        spin_lock write_queue_lock{};
        datagram_packet* dg_write_queue_head{nullptr};
        datagram_packet* dg_write_queue_tail{nullptr};

        write_ring_queue<std::weak_ptr<connection::VoiceClientConnection>, 1024 * 8> voice_write_queue{};
    };

    struct io_loop {
        std::thread base_dispatcher{};
        struct event_base* event_base{nullptr};

        std::mutex entries_mutex{};
        std::vector<io_loop_entry*> registered_entries{};
    };

    struct io_binding {
        VirtualServer* virtual_server{nullptr};
        sockaddr_storage address{};

        size_t loop_entry_index{0};
        std::vector<io_loop_entry*> loop_entries{};
    };

    enum struct ServerRegisterResult {
        SUCCESS,
        FAILED_TO_BIND
    };

    class Server {
        public:
            Server();
            ~Server();

            bool initialize(std::string& /* error */);
            void finalize();

            ServerRegisterResult register_virtual_server(VirtualServer* /* server */);
            /* this will block until all executions have been finished */
            void unregister_virtual_server(VirtualServer* /* server */);


            void schedule_client_write(const std::shared_ptr<connection::VoiceClientConnection>& /* client */);
            void unregister_client(const std::shared_ptr<connection::VoiceClientConnection>& /* client */);
        private:
            std::mutex io_lock{};
            std::vector<io_loop*> io_loops{};

            std::mutex bindings_lock{};
            std::vector<io_binding*> io_bindings{}; /* may contains nullptr! */

            std::thread client_tick_thread{};

            void execute_client_ticking();
    };
}