//
// Created by WolverinDEV on 03/03/2020.
//

#include <deque>
#include "./ClientChannelService.h"
#include "../client/ConnectedClient.h"
#include "../vserver/VirtualServer.h"
#include "../groups/GroupManager.h"
#include "../groups/GroupAssignmentManager.h"
#include <misc/timer.h>
#include <misc/sassert.h>
#include <log/LogUtils.h>

using namespace ts::server::channels;

ts::ServerId ClientChannelService::get_server_id() const {
    return this->virtual_server_->server_id();
}

ChannelGroupInheritance ClientChannelService::calculate_channel_group(ChannelId channel_id, ClientDbId client_dbid, ClientType client_type) {
    auto server_state_lock = this->virtual_server_->lock_server_status();
    auto tree_lock = this->virtual_server_->lock_channel_clients();
    tree_lock.auto_lock_shared();

    auto assignments = this->virtual_server_->group_manager()->assignments().channel_groups_of_client(groups::GroupAssignmentCalculateMode::GLOBAL, client_dbid);

    auto inheritance_channel = this->virtual_server_->server_channel_tree()->findChannel(channel_id);
    while(inheritance_channel) {
        auto inheritance_channel_id = inheritance_channel->channelId();
        auto assignment = std::find_if(assignments.begin(), assignments.end(), [&](const groups::ChannelGroupAssignment& assignment) {
            return assignment.channel_id == inheritance_channel_id;
        });

        if(assignment == assignments.end() || !assignment->group_id) {
            auto permission = inheritance_channel->permissions()->permission_value_flagged(permission::b_channel_group_inheritance_end);
            if(permission::v2::permission_granted(1, permission))
                break;

            inheritance_channel = inheritance_channel->parent();
        } else {
            return {
                assignment->group_id,
                inheritance_channel_id
            };
        }
    }

    auto default_group = this->virtual_server_->group_manager()->default_channel_group(client_type);
    assert(default_group);
    return {
        default_group->group_id(),
        channel_id
    };
}

/*
 * General notice:
 * We're locking the clients command_lock, so when changing the currentChannel variable nothing gets messed up.
 * As a little note: This must be done before locking the channel tree at all.
 */

ClientMoveResult ClientChannelService::client_ban(const std::shared_ptr<ConnectedClient> &target,
                                      const std::shared_ptr<ConnectedClient> &invoker, const std::string &reason, size_t time,
                                      ts::rwshared_lock<ts::rw_mutex> &tree_lock) {
    tree_lock.auto_unlock();
    auto command_lock = target->lock_command_handling();
    if(auto err = tree_lock.auto_lock_exclusive(); err)
        return ClientMoveResult::TREE_LOCK_FAILED;

    auto server_state_lock = this->virtual_server_->lock_server_status();
    std::unique_lock client_chan_tree_lock{target->get_channel_lock()};
    auto old_channel = target->getChannel();
    target->setChannel(nullptr);
    client_chan_tree_lock.unlock();

    ChannelId old_channel_id{0};
    if(old_channel) {
        old_channel_id = old_channel->channelId();
        for(const auto& client : this->virtual_server_->connected_clients(false)) {
            if(!client || client == target)
                continue;

            std::unique_lock client_channel_lock(client->get_channel_lock());
            if(client->isClientVisible(target, false))
                client->notifyClientLeftViewBanned(target, old_channel_id, invoker, time, false, reason);
        }

        auto s_channel = dynamic_pointer_cast<ServerChannel>(old_channel);
        assert(s_channel);
        s_channel->unregister_client(target);
    }

    target->notifyClientLeftViewBanned(target, old_channel_id, invoker, time, false, reason);
    return ClientMoveResult::SUCCESS;
}

ClientMoveResult ClientChannelService::client_kick(const std::shared_ptr<ConnectedClient> &target,
                                                   const std::shared_ptr<ConnectedClient> &invoker, const std::string &reason,
                                                   const std::shared_ptr<ServerChannel> &target_channel,
                                                   ts::rwshared_lock<ts::rw_mutex> &tree_lock) {
    if(target_channel) {
        return this->client_move(target, target_channel, invoker, reason, ViewReasonId::VREASON_CHANNEL_KICK, true, tree_lock);
    } else {
        tree_lock.auto_unlock();
        auto command_lock = target->lock_command_handling();
        if(auto err = tree_lock.auto_lock_exclusive(); err)
            return ClientMoveResult::TREE_LOCK_FAILED;

        auto server_state_lock = this->virtual_server_->lock_server_status();
        std::unique_lock client_chan_tree_lock{target->get_channel_lock()};
        auto old_channel = target->getChannel();
        target->setChannel(nullptr);
        client_chan_tree_lock.unlock();

        ChannelId old_channel_id{0};
        if(old_channel) {
            old_channel_id = old_channel->channelId();
            for(const auto& client : this->virtual_server_->connected_clients(false)) {
                if(!client || client == target)
                    continue;

                std::unique_lock client_channel_lock(client->get_channel_lock());
                if(client->isClientVisible(target, false))
                    client->notifyClientLeftViewKicked(target, old_channel_id, reason, invoker, false, nullptr);
            }

            auto s_channel = dynamic_pointer_cast<ServerChannel>(old_channel);
            assert(s_channel);
            s_channel->unregister_client(target);
        }

        target->notifyClientLeftViewKicked(target, old_channel_id, reason, invoker, false, nullptr);
    }

    return ClientMoveResult::SUCCESS;
}

ClientMoveResult ClientChannelService::client_move(const std::shared_ptr<ConnectedClient> &target,
                                                    std::shared_ptr<ServerChannel> s_target_channel,
                                                    const std::shared_ptr<ConnectedClient> &invoker,
                                                    const std::string &reason_message,
                                                    ViewReasonId reason_id,
                                                    bool notify_client,
                                                    ts::rwshared_lock<ts::rw_mutex> &tree_lock) {
    TIMING_START(timings);
    tree_lock.auto_unlock();

    auto command_lock = target->lock_command_handling();
    if(auto err = tree_lock.auto_lock_exclusive(); err)
        return ClientMoveResult::TREE_LOCK_FAILED;

    auto server_state_lock = this->virtual_server_->lock_server_status();

    TIMING_STEP(timings, "tree lock setup");
    if(target->getChannel() == s_target_channel)
        return ClientMoveResult::SUCCESS;

    /* first step: resolve the target channel / or fix missing */
    auto s_source_channel = dynamic_pointer_cast<ServerChannel>(target->getChannel());
    assert(!target->getChannel() || s_source_channel != nullptr);

    auto server_channel_tree = this->virtual_server_->server_channel_tree();
    std::deque<property::ClientProperties> client_updates;
    std::deque<property::ClientProperties> changed_groups{};
    if(s_target_channel) {
        /* don't let the client join a deleted channel */
        if(s_target_channel->deleted) {
            s_target_channel = dynamic_pointer_cast<ServerChannel>(server_channel_tree->getDefaultChannel());
            assert(s_target_channel);
        }

        /* update the group properties here already, so for all enter views we could just send the new props directly */
        changed_groups = this->virtual_server_->group_manager()->assignments().update_client_group_properties(target, s_target_channel->channelId());
        client_updates.insert(client_updates.end(), changed_groups.begin(), changed_groups.end()); //TODO: Only update for clients which have no enter?
    }

    auto l_target_channel = s_target_channel ? server_channel_tree->findLinkedChannel(s_target_channel->channelId()) : nullptr;
    auto l_source_channel = s_source_channel ? server_channel_tree->findLinkedChannel(s_source_channel->channelId()) : nullptr;
    TIMING_STEP(timings, "channel lookup ");


    /* second step: show the target channel to the client if its not shown and let him subscibe to the channel */
    if(s_target_channel && notify_client) {
        std::unique_lock client_channel_lock(target->get_channel_lock());

        bool success = false;
        /* TODO: Use a bunk here and not a notify for every single */
        for(const auto& channel : target->channel_view()->show_channel(l_target_channel, success))
            target->notifyChannelShow(channel->channel(), channel->previous_channel);
        sassert(success);
        if(!success)
            return ClientMoveResult::VIEW_INSERT_FAILED;

        target->subscribeChannel({s_target_channel}, false, true);
    }
    TIMING_STEP(timings, "target show chan");

    if(s_target_channel) {
        for(const auto& client : this->virtual_server_->connected_clients(false)) {
            if (!notify_client && client == target) continue;

            std::unique_lock client_channel_lock{client->get_channel_lock()};
            auto chan_target = client->channel_view()->find_channel(s_target_channel);

            if(chan_target) {
                auto chan_source = client->channel_view()->find_channel(s_source_channel);
                if(chan_source) {
                    if (chan_target->subscribed || client == target) {
                        if (client == target || client->isClientVisible(target, false)) {
                            client->notifyClientMoved(target, s_target_channel, reason_id, reason_message, invoker, false);
                        } else {
                            client->notifyClientEnterView(target, invoker, reason_message, s_target_channel, reason_id, s_source_channel, false);
                        }
                    } else if(client->isClientVisible(target, false)){
                        //Client got out of view
                        client->notifyClientLeftView(target, s_target_channel, reason_id, reason_message.empty() ? std::string{"view left"} : reason_message, invoker, false);
                    }
                } else {
                    if(client == target && client->getType() != ClientType::CLIENT_INTERNAL && client->getType() != ClientType::CLIENT_MUSIC)
                        logCritical(this->get_server_id(), "{} Client enters visibility twice!", CLIENT_STR_LOG_PREFIX_(client));

                    //Client entered view
                    if(chan_target->subscribed)
                        client->notifyClientEnterView(target, invoker, reason_message, s_target_channel, ViewReasonId::VREASON_USER_ACTION, nullptr, false);
                }
            } else {
                /* target channel isn't visible => so client gone out of view */
                if(client == target && client->getType() != ClientType::CLIENT_INTERNAL && client->getType() != ClientType::CLIENT_MUSIC)
                    logCritical(this->get_server_id(), "{} Moving own client into a not visible channel! This shall not happen!", CLIENT_STR_LOG_PREFIX_(client));
                //Test for in view? (Notify already does but nvm)

                if(client->isClientVisible(target, false)){
                    //Client got out of view
                    if(reason_id == ViewReasonId::VREASON_USER_ACTION)
                        client->notifyClientLeftView(target, nullptr, ViewReasonId::VREASON_SERVER_LEFT, reason_message.empty() ? "joined a hidden channel" : reason_message, invoker, false);
                    else
                        client->notifyClientLeftView(target, nullptr, ViewReasonId::VREASON_SERVER_LEFT, reason_message.empty() ? "moved to a hidden channel" : reason_message, invoker, false);
                }
            }

        }

        if(s_source_channel)
            s_source_channel->unregister_client(target);
        s_target_channel->register_client(target);
    } else {
        /* client left the server */
        if(target->getChannel()) {
            for(const auto& client : this->virtual_server_->connected_clients(false)) {
                if(!client || client == target)
                    continue;

                std::unique_lock client_channel_lock(client->get_channel_lock());
                if(client->isClientVisible(target, false))
                    client->notifyClientLeftView(target, nullptr, reason_id, reason_message, invoker, false);
            }

            s_source_channel->unregister_client(target);
        }
    }
    TIMING_STEP(timings, "notify viewers ");

    target->setChannel(s_target_channel);
    tree_lock.downgrade_lock();

    std::unique_lock client_channel_lock(target->get_channel_lock());
    TIMING_STEP(timings, "lock own ch tree");

    if (s_source_channel) {
        s_source_channel->properties()[property::CHANNEL_LAST_LEFT] = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::system_clock::now().time_since_epoch()).count();

        this->virtual_server_->group_manager()->assignments().cleanup_channel_temporary_assignment(target->getClientDatabaseId(), s_source_channel->channelId());
        auto update = target->properties()[property::CLIENT_IS_TALKER].as<bool>() || target->properties()[property::CLIENT_TALK_REQUEST].as<int64_t>() > 0;

        if(update) {
            target->properties()[property::CLIENT_IS_TALKER] = 0;
            target->properties()[property::CLIENT_TALK_REQUEST] = 0;
            target->properties()[property::CLIENT_TALK_REQUEST_MSG] = "";
            client_updates.push_back(property::CLIENT_IS_TALKER);
            client_updates.push_back(property::CLIENT_TALK_REQUEST);
            client_updates.push_back(property::CLIENT_TALK_REQUEST_MSG);
        }
        TIMING_STEP(timings, "src chan up");
    }

    if (s_target_channel) {
        if(target->update_cached_permissions()) /* update cached calculated permissions */
            target->sendNeededPermissions(false);
        TIMING_STEP(timings, "perm gr upd");

        if(s_source_channel) {
            std::deque<ChannelId> deleted;
            for(const auto& channel : target->channel_view()->test_channel(l_source_channel, l_target_channel))
                deleted.push_back(channel->channelId());

            if(!deleted.empty())
                target->notifyChannelHide(deleted, false);

            /* force unsubscribe from old channel */
            auto i_source_channel = s_source_channel->channelId();
            if(std::find(deleted.begin(), deleted.end(), i_source_channel) == deleted.end()) {
                auto source_channel_sub_power = target->calculate_permission(permission::i_channel_subscribe_power, i_source_channel);
                if(!s_source_channel->permission_granted(permission::i_channel_needed_subscribe_power, source_channel_sub_power, false)) {
                    auto source_channel_sub_power_ignore = target->calculate_permission(permission::b_channel_ignore_subscribe_power, i_source_channel);
                    if(!permission::v2::permission_granted(1, source_channel_sub_power_ignore, true)) {
                        logTrace(this->get_server_id(), "Force unsubscribing of client {} for channel {}/{}. (Channel switch and no permissions)",
                                 CLIENT_STR_LOG_PREFIX_(target), s_source_channel->name(),
                                 i_source_channel
                        );
                        target->unsubscribeChannel({s_source_channel}, false); //Unsubscribe last channel (hasn't permissions)
                    }
                }
            }
            TIMING_STEP(timings, "src hide ts");
        }
    }
    client_channel_lock.unlock();
    /* both methods lock if they require stuff */
    this->virtual_server_->broadcast_service().client_updated(target, client_updates, s_source_channel ? true : false);
    TIMING_STEP(timings, "notify cpro");
    if(s_target_channel) {
        target->updateChannelClientProperties(false, s_source_channel ? true : false);
        TIMING_STEP(timings, "notify_t_pr");
    }
    debugMessage(this->get_server_id(), "{} Client move timings: {}", CLIENT_STR_LOG_PREFIX_(target), TIMING_FINISH(timings));
    return ClientMoveResult ::SUCCESS;
}

/*
 * 1. flag channel as deleted (lock channel tree so no moves)
 * 2. Gather all clients within the channel (lock their execute lock)
 * 3. Unlock channel tree and lock client locks
 * 4. lock channel tree again and move the clients (No new clients should be joined because channel is flagged as deleted!)
 *
 * Note: channel cant be a ref because the channel itself gets deleted!
 */
ChannelDeleteResult ClientChannelService::delete_channel(std::shared_ptr<ServerChannel> channel, const std::shared_ptr<ConnectedClient> &invoker, const std::string &kick_message, ts::rwshared_lock<ts::rw_mutex> &tree_lock) {
    if(auto err = tree_lock.auto_lock_exclusive(); err)
        return ChannelDeleteResult::TREE_LOCK_FAILED;

    if(channel->deleted)
        return ChannelDeleteResult::SUCCESS;

    std::deque<std::shared_ptr<ConnectedClient>> clients;
    {
        for(const auto& sub_channel : this->virtual_server_->server_channel_tree()->channels(channel)) {
            auto s_channel = dynamic_pointer_cast<ServerChannel>(sub_channel);
            assert(s_channel);

            auto chan_clients = this->virtual_server_->find_clients_by_channel(s_channel);
            clients.insert(clients.end(), chan_clients.begin(), chan_clients.end());
            s_channel->deleted = true;
        }
        auto chan_clients = this->virtual_server_->find_clients_by_channel(channel);
        clients.insert(clients.end(), chan_clients.begin(), chan_clients.end());
        channel->deleted = true;
    }
    auto default_channel = dynamic_pointer_cast<ServerChannel>(this->virtual_server_->server_channel_tree()->getDefaultChannel());
    assert(default_channel);
    tree_lock.auto_unlock();

    std::deque<std::unique_lock<threads::Mutex>> command_locks;
    for(const auto& client : clients)
        command_locks.push_back(std::move(client->lock_command_handling<std::unique_lock<threads::Mutex>>()));

    for(const auto& client : clients) {
        auto result = this->client_move(client, default_channel, invoker, kick_message, ViewReasonId::VREASON_CHANNEL_KICK, true, tree_lock);
        if(result != ClientMoveResult::SUCCESS)
            return ChannelDeleteResult::CLIENT_MOVE_FAILED;
    }

    if(auto err = tree_lock.auto_lock_exclusive(); err)
        return ChannelDeleteResult::TREE_LOCK_FAILED;
    command_locks.clear();

    auto channel_ids = this->virtual_server_->server_channel_tree()->delete_channel_root(channel);
    tree_lock.auto_lock_shared();

    for(auto& client : this->virtual_server_->connected_clients(false)) {
        std::unique_lock client_channel_lock(client->get_channel_lock());
        client->notifyChannelDeleted(client->channel_view()->delete_channel_root(channel), invoker);
    }
    return ChannelDeleteResult::SUCCESS;
}