//
// Created by WolverinDEV on 03/03/2020.
//

#include "./VirtualServerBase.h"
#include "./VirtualServerBase.h"
#include "../client/ConnectedClient.h"
#include "VirtualServerBase.h"
#include <log/LogUtils.h>
#include <src/InstanceHandler.h>

using namespace ts::server::vserver;

bool VirtualServerBase::initialize(std::string &error) {
    //TODO?
    return true;
}

VirtualServerStartResult VirtualServerBase::start_server_virtual(std::string &error) {
    ts::rwshared_lock slock{this->status_mutex_, ts::rw_lock_exclusive};
    if(this->status_ != VirtualServerStatus::STOPPED)
        return VirtualServerStartResult::SERVER_IS_NOT_OFFLINE;
    slock.downgrade_lock();

    if(auto err = this->start_server_virtual_(slock, error); err != VirtualServerStartResult::SUCCESS)
        return err;

    if(auto err = slock.upgrade_lock(); err)
        logCritical(this->server_id(), "Failed to lock status lock after server virtual start. Ignoring it and still setting state to virtual!");

    this->status_ = VirtualServerStatus::VIRTUAL;
    return VirtualServerStartResult::SUCCESS;
}

VirtualServerStartResult VirtualServerBase::start_server(std::string &error) {
    ts::rwshared_lock slock{this->status_mutex_, ts::rw_lock_exclusive};

    if(this->status_ != VirtualServerStatus::VIRTUAL && this->status_ == VirtualServerStatus::STOPPED)
        return VirtualServerStartResult::SERVER_IS_NOT_OFFLINE;

    auto full_start = this->status_ == VirtualServerStatus::STOPPED;
    this->status_ = VirtualServerStatus::BOOTING;

    /* even if queries could be connected, a start from virtual should not lead to a crash! */
    slock.downgrade_lock();

    if(full_start)
        if(auto err = this->start_server_virtual_(slock, error); err != VirtualServerStartResult::SUCCESS)
            return err;

    if(auto err = this->start_server_(slock, error); err != VirtualServerStartResult::SUCCESS)
        return err;

    if(auto err = slock.upgrade_lock(); err)
        logCritical(this->server_id(), "Failed to lock status lock after server start. Ignoring it and still setting state to running!");

    this->status_ = VirtualServerStatus::RUNNING;
    return VirtualServerStartResult::SUCCESS;
}

VirtualServerStopResult VirtualServerBase::stop_server() {
    ts::rwshared_lock slock{this->status_mutex_, ts::rw_lock_shared};
    if(this->status_ != VirtualServerStatus::RUNNING && this->status_ != VirtualServerStatus::VIRTUAL)
        return VirtualServerStopResult::SERVER_IS_NOT_RUNNING;

    if(auto err = slock.upgrade_lock(); err)
        return VirtualServerStopResult::STATE_LOCK_FAILED;

    this->status_ = VirtualServerStatus::STOPPING;
    slock.downgrade_lock();
    this->stop_server_(slock);

    if(auto err = slock.upgrade_lock(); err)
        logCritical(this->server_id(), "Failed to lock status lock after server stop. Ignoring it and still setting state to stopped!");

    this->status_ = VirtualServerStatus::STOPPED;
    return VirtualServerStopResult::SUCCESS;
}

VirtualServerStartResult VirtualServerBase::start_server_(ts::rwshared_lock<ts::rw_mutex> &slock, std::string &error) {
    //TODO: Load server permissions, etc

    error = "Not implemented";
    return VirtualServerStartResult::CUSTOM;
}

VirtualServerClientRegisterResult VirtualServerBase::register_client(const std::shared_ptr<ConnectedClient> &client) {
    assert(client);

    ts::rwshared_lock status_lock{this->status_mutex_};
    if(this->server_status() != VirtualServerStatus::VIRTUAL && this->server_status() != VirtualServerStatus::RUNNING)
    return VirtualServerClientRegisterResult::SERVER_NOT_ONLINE;

    {
        std::lock_guard client_lock{this->registered_clients.mutex};
        if(client->getClientId() > 0) {
            logCritical(this->server_id(), "Client {} ({}|{}) has been already registered somewhere!", client->getDisplayName(), client->getClientId(), client->getUid());
            return VirtualServerClientRegisterResult::CLIENT_ALREADY_KNOWN;
        }

        ClientId client_id = 0;
        ClientId max_client_id = this->registered_clients.clients.size();

        while(client_id < max_client_id && this->registered_clients.clients[client_id])
            client_id++;

        if(client_id > 1024 * 8)
            return VirtualServerClientRegisterResult::PROTOCOL_LIMIT_REACHED;

        if(client_id == max_client_id)
            this->registered_clients.clients.push_back(client);
        else
            this->registered_clients.clients[client_id] = client;
        this->registered_clients.count++;
        client->setClientId(client_id);
    }

    {
        std::lock_guard nlock{this->client_nickname_lock_};

        auto login_name = client->getDisplayName();
        while(login_name.length() < 3)
            login_name += ".";

        if(client->getExternalType() == ClientType::CLIENT_TEAMSPEAK)
            client->properties()[property::CLIENT_LOGIN_NAME] = login_name;

        std::shared_ptr<ConnectedClient> found_client = nullptr;

        auto client_name = login_name;
        size_t counter = 0;

        {
            std::lock_guard clients_lock(this->registered_clients.mutex);
            while(true) {
                for(auto& _client : this->registered_clients.clients) {
                    if(!_client) continue;

                    if(_client->getDisplayName() == client_name && _client != client)
                        goto increase_name;
                }
                goto nickname_valid;

                increase_name:
                client_name = login_name + std::to_string(++counter);
            }
        }
        nickname_valid:
        client->setDisplayName(client_name);
    }

    {
        auto properties = this->server_properties();
        if(client->getType() == ClientType::CLIENT_TEAMSPEAK || client->getType() == ClientType::CLIENT_WEB) {
            properties->find(property::VIRTUALSERVER_CLIENT_CONNECTIONS) ++; //increase client connections
            properties->find(property::VIRTUALSERVER_LAST_CLIENT_CONNECT) = std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now().time_since_epoch()).count();
        } else if(client->getType() == ClientType::CLIENT_QUERY) {
            properties->find(property::VIRTUALSERVER_LAST_QUERY_CONNECT) ++; //increase query connections
            properties->find(property::VIRTUALSERVER_QUERY_CLIENT_CONNECTIONS) = std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now().time_since_epoch()).count();
        }
    }

    return VirtualServerClientRegisterResult::SUCCESS;
}

void VirtualServerBase::unregister_client(const std::shared_ptr<ConnectedClient> &client) {
    ts::rwshared_lock status_lock{this->status_mutex_, ts::rw_lock_defered};
    {
        auto execute_lock = client->lock_command_handling();
        if(client->getChannel()) {
            auto tree_lock = this->lock_channel_clients();
            auto result = this->channel_service().client_move(client, nullptr, nullptr, "", ViewReasonId::VREASON_SERVER_LEFT, false, tree_lock);
            if(result != channels::ClientMoveResult::SUCCESS)
                logCritical(this->server_id(), "Failed to unregister client from channel tree on client unregister ({}).", (int) result);
        }

        status_lock.auto_lock_shared();

        auto client_id = client->getClientId();
        if(client_id == 0) return; /* client hasn't been registered */

        {
            std::lock_guard lock(this->registered_clients.mutex);
            if(client_id >= this->registered_clients.clients.size()) {
                logCritical(this->server_id(), "Client {} ({}|{}) has been registered, but client id exceed client id! Failed to unregister client.", client->getDisplayName(), client_id, client->getUid());
            } else {
                auto& client_container = this->registered_clients.clients[client_id];
                if(client_container != client) {
                    logCritical(this->server_id(), "Client {} ({}|{}) has been registered, but container hasn't client set! Failed to unregister client.", client->getDisplayName(), client_id, client->getUid());
                } else {
                    client_container.reset();
                    this->registered_clients.count--;
                }
            }
        }
    }

    {
        auto properties = this->server_properties();
        if(client->getType() == ClientType::CLIENT_TEAMSPEAK || client->getType() == ClientType::CLIENT_WEB) {
            properties->find(property::VIRTUALSERVER_LAST_CLIENT_DISCONNECT) = std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now().time_since_epoch()).count();
        } else if(client->getType() == ClientType::CLIENT_QUERY) {
            properties->find(property::VIRTUALSERVER_LAST_QUERY_DISCONNECT) = std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now().time_since_epoch()).count();
        }
    }

    {
        auto channel_clients_lock = this->lock_channel_clients();
        if(client->getChannel()) {//We dont have to make him invisible if he hasn't even a channel
            auto result = this->channel_service().client_move(client, nullptr, nullptr, "", ViewReasonId::VREASON_SERVER_LEFT, false, channel_clients_lock);;
            if(result != channels::ClientMoveResult::SUCCESS)
                logCritical(this->server_id(), "Failed to unregister client {} from the server channel tree ({}).", client->getDisplayName(), (int) result);
        }
    }

    serverInstance->databaseHelper()->saveClientPermissions(this->server_id(), client->getClientDatabaseId(), client->permissions());
    client->setClientId(0);
}