#pragma once

#include <Definitions.h>
#include <lock/rw_mutex.h>
#include <PermissionRegister.h>
#include <sql/SqlQuery.h>
#include <src/services/VirtualServerBroadcastService.h>
#include "../services/PermissionsService.h"
#include "../services/ClientChannelService.h"

namespace ts::server {
    class ConnectedClient;

    namespace groups {
        class GroupManager;
    }
}

namespace ts {
    class Properties;
    class ServerChannelTree;

    namespace stats {
        class ConnectionStatistics;
    }
}


namespace ts::server::vserver {
    struct VirtualServerStatus {
        enum value {
            STOPPED, /* no data has been loaded */

            BOOTING, /* loading & initializing data */
            VIRTUAL, /* only available for query clients */

            STARTING, /* start bindings */
            RUNNING, /* server is running */

            STOPPING, /* server is stopping and unloading any data */

            DEPLOYING /* server is in deploy modus */
        };
    };

    enum struct VirtualServerStartResult {
        SUCCESS,
        BIND_FAILED,
        SERVER_IS_DEFAULT,
        CUSTOM,

        SERVER_IS_NOT_OFFLINE
    };

    enum struct VirtualServerStopResult {
        SUCCESS,
        SERVER_IS_NOT_RUNNING,
        STATE_LOCK_FAILED
    };

    enum struct VirtualServerClientRegisterResult {
        SUCCESS,
        PROTOCOL_LIMIT_REACHED,
        CLIENT_ALREADY_KNOWN,
        SERVER_NOT_ONLINE
    };

#define var_access_requires_server_loaded(variable)         \
do {                                                        \
    assert(this->status_ != VirtualServerStatus::STOPPED);  \
    return variable;                                        \
} while(0);

    class VirtualServerBase {
        public:
            //TODO: Constructor
            //VirtualServerBase();
            ~VirtualServerBase();

            [[nodiscard]] inline std::shared_ptr<VirtualServerBase> server_ref() { return this->server_ref_.lock(); }

            /* basic initialize to the STOPPED state */
            virtual bool initialize(std::string& /* error */);
            VirtualServerStartResult start_server_virtual(std::string& /* error */);
            VirtualServerStartResult start_server(std::string& /* error */);
            VirtualServerStopResult stop_server();

            /* basic information */
            [[nodiscard]] inline ServerId server_id() const { return this->server_id_; }
            [[nodiscard]] inline VirtualServerStatus::value server_status() const { return this->status_; }
            [[nodiscard]] inline ts::rwshared_lock<ts::rw_mutex> lock_server_status() { return ts::rwshared_lock{this->status_mutex_, ts::rw_lock_shared}; }

            [[nodiscard]] inline sql::SqlManager* sql_manager() { return this->sql_manager_; }
            [[nodiscard]] inline const std::shared_ptr<stats::ConnectionStatistics>& server_connection_statistics() { return this->server_connection_statistics_; }

            [[nodiscard]] inline permissions::PermissionService& permission_service() { return this->permission_service_; }
            [[nodiscard]] inline BroadcastService& broadcast_service() { return this->broadcast_service_; }
            
            [[nodiscard]] inline channels::ClientChannelService&  channel_service() { return this->channel_service_; }
            [[nodiscard]] inline ts::rwshared_lock<ts::rw_mutex> lock_channel_clients() { return ts::rwshared_lock{this->channel_clients_lock_, ts::rw_lock_shared}; }
            [[nodiscard]] inline std::mutex& client_nickname_lock() { return this->client_nickname_lock_; }

            [[nodiscard]] inline const std::shared_ptr<Properties>& server_properties() { var_access_requires_server_loaded(this->server_properties_); }
            [[nodiscard]] inline const std::shared_ptr<ServerChannelTree>& server_channel_tree() { var_access_requires_server_loaded(this->server_channel_tree_); }

            [[nodiscard]] inline const std::shared_ptr<groups::GroupManager>& group_manager() { var_access_requires_server_loaded(this->group_manager_); }

            /* client management */
            /* this may block until the current command of the client has been finished executed */
            void unregister_client(const std::shared_ptr<ConnectedClient>& /* client */);
            VirtualServerClientRegisterResult register_client(const std::shared_ptr<ConnectedClient>& /* client */);

            /* ATTENTION: lock_channel_clients should be acquired for the following methods. Else the client might be disconnected while working with him.   */
            [[nodiscard]] std::vector<std::shared_ptr<ConnectedClient>> connected_clients(bool /* must be registered within the channel tree */);
            [[nodiscard]] std::shared_ptr<ConnectedClient> find_client_by_id(ClientId /* id */);
            [[nodiscard]] std::vector<std::shared_ptr<ConnectedClient>> find_clients_by_uid(ClientId /* id */);
            [[nodiscard]] std::vector<std::shared_ptr<ConnectedClient>> find_clients_by_channel(const std::shared_ptr<ServerChannel>& /* channel */);
        protected:
            std::weak_ptr<VirtualServerBase> server_ref_{};

            sql::SqlManager* sql_manager_{nullptr}; //TODO!

            ServerId server_id_{0};
            ts::rw_mutex status_mutex_{};
            VirtualServerStatus::value status_{VirtualServerStatus::STOPPED};

            std::shared_ptr<Properties> server_properties_{nullptr};
            std::shared_ptr<ServerChannelTree> server_channel_tree_{nullptr};

            /*
             * Lock whenever we want to read/write from the channel tree or the clients within
             * Attention: We're not allowed to do permission calculations while having the channel tree in write lock!
             */
            ts::rw_mutex channel_clients_lock_{};
            std::mutex client_nickname_lock_{};

            /* some services */
            permissions::PermissionService permission_service_;
            channels::ClientChannelService channel_service_;
            BroadcastService broadcast_service_;

            struct {
                size_t count{0};
                std::mutex mutex{};
                std::vector<std::shared_ptr<ConnectedClient>> clients{};
            } registered_clients;

            std::shared_ptr<groups::GroupManager> group_manager_;

            std::shared_ptr<stats::ConnectionStatistics> server_connection_statistics_;

            /* private methods */
            virtual VirtualServerStartResult start_server_virtual_(ts::rwshared_lock<ts::rw_mutex>& /* state lock */, std::string& /* error */);
            virtual VirtualServerStartResult start_server_(ts::rwshared_lock<ts::rw_mutex>& /* state lock */, std::string& /* error */);
            virtual void stop_server_(ts::rwshared_lock<ts::rw_mutex>& /* state lock */);

            virtual void execute_tick(const std::chrono::system_clock::time_point& /* now */) = 0;
    };
}